# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the sdss directory in idlutils.
"""
import os
import re
import numpy as np
from astropy.io import fits
from astropy.utils.data import download_file
from . import PydlutilsException
from .spheregroup import spherematch
from .yanny import yanny
from .. import uniq
#
# Cache for the maskbits file.
#
maskbits = None
#
# Cache the sweep index
#
sweep_cache = {'star': None, 'gal': None, 'sky': None}
#
# Cache sdss_astrombad data
#
opbadfields = None


def default_skyversion():
    """Returns skyversion number to use for photoop imaging.

    Returns
    -------
    :class:`int`
        The default skyversion number.

    Notes
    -----
    The skyversion number is hard-coded to 2.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import default_skyversion
    >>> default_skyversion()
    2
    """
    return 2


def sdss_astrombad(run, camcol, field, photolog_version='dr10'):
    """For a list of RUN, CAMCOL, FIELD, return whether each field has bad
    astrometry.

    Parameters
    ----------
    run, camcol, field : :class:`int` or array of :class:`int`
        Run, camcol and field.  If arrays are passed,
        all must have the same length.
    photolog_version : :class:`str`, optional
        Use this version of photolog to obtain the obBadfields.par file,
        if :envvar:`PHOTOLOG_DIR` is not set.

    Returns
    -------
    :class:`numpy.ndarray` of :class:`bool`
        Array of bool.  ``True`` indicates the field is bad.

    Raises
    ------
    :exc:`ValueError`
        If the sizes of the arrays don't match or if the array values are out
        of bounds.

    Notes
    -----
    Reads data from ``$PHOTOLOG_DIR/opfiles/opBadFields.par``.

    If there is a problem with one camcol, we assume a
    problem with all camcols.

    """
    global opbadfields
    #
    # Check inputs
    #
    if isinstance(run, int):
        #
        # Assume all inputs are integers & promote to arrays.
        #
        run = np.array([run], dtype=np.int64)
        camcol = np.array([camcol], dtype=np.int64)
        field = np.array([field], dtype=np.int64)
    else:
        #
        # Check that all inputs have the same shape.
        #
        if run.shape != camcol.shape:
            raise ValueError("camcol.shape does not match run.shape!")
        if run.shape != field.shape:
            raise ValueError("field.shape does not match run.shape!")
    #
    # Check ranges of parameters
    #
    if ((run < 0) | (run >= 2**16)).any():
        raise ValueError("run values are out-of-bounds!")
    if ((camcol < 1) | (camcol > 6)).any():
        raise ValueError("camcol values are out-of-bounds!")
    if ((field < 0) | (field >= 2**12)).any():
        raise ValueError("camcol values are out-of-bounds!")
    #
    # Read the file
    #
    if opbadfields is None:  # pragma: no cover
        if os.getenv('PHOTOLOG_DIR') is None:
            if (photolog_version == 'trunk' or
                    photolog_version.startswith('branches/')):
                iversion = photolog_version
            else:
                iversion = 'tags/'+photolog_version
            baseurl = ('https://svn.sdss.org/public/data/sdss/photolog/' +
                       '{0}/opfiles/opBadfields.par').format(iversion)
            filename = download_file(baseurl, cache=True)
        else:
            filename = os.path.join(os.getenv('PHOTOLOG_DIR'), 'opfiles',
                                    'opBadfields.par')
        astrombadfile = yanny(filename)
        w = ((astrombadfile['BADFIELDS']['problem'] == 'astrom'.encode()) |
             (astrombadfile['BADFIELDS']['problem'] == 'rotator'.encode()))
        opbadfields = astrombadfile['BADFIELDS'][w]
    #
    # opbadfields already has astrom problems selected at this point
    #
    bad = np.zeros(run.shape, dtype=bool)
    for row in opbadfields:
        w = ((run == row['run']) &
             (field >= row['firstfield']) & (field < row['lastfield']))
        if w.any():
            bad[w] = True
    return bad


def sdss_flagexist(flagname, bitname, flagexist=False, whichexist=False):
    """Check for the existence of flags.

    Parameters
    ----------
    flagname : :class:`str`
        The name of a bitmask group. Not case-sensitive.
    bitname : :class:`str` or :class:`list`
        The name(s) of the specific bitmask(s) within the `flagname` group.
    flagexist : :class:`bool`, optional
        If flagexist is True, return a tuple with the second component
        indicating whether the binary flag named `flagname` exists, even
        if `bitname` is wrong.
    whichexist : :class:`bool`, optional
        If whichexist is True, return a list containing existence test results
        for each individual flag.

    Returns
    -------
    :class:`bool` or :class:`tuple`
        A boolean value or a tuple of bool.
    """
    global maskbits
    if maskbits is None:  # pragma: no cover
        maskbits = set_maskbits()
    #
    # Make sure label is a list
    #
    if isinstance(bitname, (str,)):
        bitnames = [bitname.upper()]
    else:
        bitnames = [b.upper() for b in bitname]
    f = False
    l = False
    which = [False] * len(bitnames)
    if flagname.upper() in maskbits:
        f = True
        which = [n in maskbits[flagname.upper()] for n in bitnames]
        l = sum(which) == len(which)
    if flagexist and whichexist:
        return (l, f, which)
    elif flagexist:
        return (l, f)
    elif whichexist:
        return (l, which)
    else:
        return l


def sdss_flagname(flagname, flagvalue, concat=False):
    """Return a list of flag names corresponding to the values.

    Parameters
    ----------
    flagname : :class:`str`
        The name of a bitmask group. Not case-sensitive.
    flagvalue : :class:`long`
        The value to be converted into bitmask names.
    concat : :class:`bool`, optional
        If set to ``True``, the list of names is converted to a
        space-separated string.

    Returns
    -------
    :class:`str` or :class:`list`
        The names of the bitmasks encoded in `flagvalue`.

    Raises
    ------
    :exc:`KeyError`
        If `flagname` is invalid

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_flagname
    >>> sdss_flagname('ANCILLARY_TARGET1',2310346608843161600) # doctest: +REMOTE_DATA
    ['BRIGHTGAL', 'BLAZGX', 'ELG']
    """
    global maskbits
    if maskbits is None:  # pragma: no cover
        maskbits = set_maskbits()
    flagu = flagname.upper()
    flagvaluint = np.uint64(flagvalue)
    one = np.uint64(1)
    bits = [bit for bit in range(64)
            if (flagvaluint & (one << np.uint64(bit))) != 0]
    retval = list()
    for bit in bits:
        try:
            f = [x for x in maskbits[flagu].items() if x[1] == bit]
        except KeyError:
            raise KeyError("Unknown flag group {0}!".format(flagu))
        if f:
            retval.append(f[0][0])
    if concat:
        retval = ' '.join(retval)
    return retval


def sdss_flagval(flagname, bitname):
    """Convert bitmask names into values.

    Converts human-readable bitmask names into numerical values.  The inputs
    are not case-sensitive; all inputs are converted to upper case internally.

    Parameters
    ----------
    flagname : :class:`str`
        The name of a bitmask group.
    bitname : :class:`str` or :class:`list`
        The name(s) of the specific bitmask(s) within the `flagname` group.

    Returns
    -------
    :class:`numpy.uint64`
        The value of the bitmask name(s).

    Raises
    ------
    :exc:`KeyError`
        If `flagname` or `bitname` are invalid names.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_flagval
    >>> sdss_flagval('ANCILLARY_TARGET1',['BLAZGX','ELG','BRIGHTGAL']) # doctest: +REMOTE_DATA
    2310346608843161600
    """
    global maskbits
    if maskbits is None:  # pragma: no cover
        maskbits = set_maskbits()
    #
    # Make sure inlabel is a list
    #
    if isinstance(bitname, (str,)):
        bitnames = [bitname.upper()]
    else:
        bitnames = [b.upper() for b in bitname]
    flagu = flagname.upper()
    flagvalue = np.uint64(0)
    for bit in bitnames:
        if flagu in maskbits:
            if bit in maskbits[flagu]:
                flagvalue += np.uint64(2)**np.uint64(maskbits[flagu][bit])
            else:
                raise KeyError("Unknown bit label {0} for flag group {1}!".format(bit, flagu))
        else:
            raise KeyError("Unknown flag group {0}!".format(flagu))
    return flagvalue


def sdss_objid(run, camcol, field, objnum, rerun=301, skyversion=None,
               firstfield=None):
    """Convert SDSS photometric identifiers into CAS-style ObjID.

    Bits are assigned in ObjID thus:

    ===== ========== ===============================================
    Bits  Name       Comment
    ===== ========== ===============================================
    63    empty      unassigned
    59-62 skyVersion resolved sky version (0-15)
    48-58 rerun      number of pipeline rerun
    32-47 run        run number
    29-31 camcol     camera column (1-6)
    28    firstField is this the first field in segment? Usually 0.
    16-27 field      field number within run
    0-15  object     object number within field
    ===== ========== ===============================================

    Parameters
    ----------
    run, camcol, field, objnum : :class:`int` or array of int
        Run, camcol, field and object number within field.  If arrays are
        passed, all must have the same length.
    rerun, skyversion, firstfield : :class:`int` or array of int, optional
        `rerun`, `skyversion` and `firstfield` usually don't change at all,
        especially for ObjIDs in DR8 and later.  If supplied,
        make sure the size matches all the other values.

    Returns
    -------
    :class:`numpy.ndarray` of :class:`numpy.int64`
        The ObjIDs of the objects.

    Raises
    ------
    :exc:`ValueError`
        If the sizes of the arrays don't match or if the array values are
        out of bounds.

    Notes
    -----
    * The ``firstField`` flag is never set in ObjIDs from DR8 and later.
    * On 32-bit systems, makes sure to explicitly declare all inputs as
      64-bit integers.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_objid
    >>> print(sdss_objid(3704,3,91,146))
    [1237661382772195474]
    """
    if skyversion is None:
        skyversion = default_skyversion()
    if firstfield is None:
        firstfield = 0
    if isinstance(run, int):
        run = np.array([run], dtype=np.int64)
    if isinstance(camcol, int):
        camcol = np.array([camcol], dtype=np.int64)
    if isinstance(field, int):
        field = np.array([field], dtype=np.int64)
    if isinstance(objnum, int):
        objnum = np.array([objnum], dtype=np.int64)
    if isinstance(rerun, int):
        if rerun == 301:
            rerun = np.zeros(run.shape, dtype=np.int64) + 301
        else:
            rerun = np.array([rerun], dtype=np.int64)
    if isinstance(skyversion, int):
        if skyversion == default_skyversion():
            skyversion = np.zeros(run.shape, dtype=np.int64) + default_skyversion()
        else:
            skyversion = np.array([skyversion], dtype=np.int64)
    if isinstance(firstfield, int):
        if firstfield == 0:
            firstfield = np.zeros(run.shape, dtype=np.int64)
        else:
            firstfield = np.array([firstfield], dtype=np.int64)

    #
    # Check that all inputs have the same shape.
    #
    if run.shape != camcol.shape:
        raise ValueError("camcol.shape does not match run.shape!")
    if run.shape != field.shape:
        raise ValueError("field.shape does not match run.shape!")
    if run.shape != objnum.shape:
        raise ValueError("objnum.shape does not match run.shape!")
    if run.shape != rerun.shape:
        raise ValueError("rerun.shape does not match run.shape!")
    if run.shape != skyversion.shape:
        raise ValueError("skyversion.shape does not match run.shape!")
    if run.shape != firstfield.shape:
        raise ValueError("firstfield.shape does not match run.shape!")
    #
    # Check ranges of parameters
    #
    if ((firstfield < 0) | (firstfield > 1)).any():
        raise ValueError("firstfield values are out-of-bounds!")
    if ((skyversion < 0) | (skyversion >= 16)).any():
        raise ValueError("skyversion values are out-of-bounds!")
    if ((rerun < 0) | (rerun >= 2**11)).any():
        raise ValueError("rerun values are out-of-bounds!")
    if ((run < 0) | (run >= 2**16)).any():
        raise ValueError("run values are out-of-bounds!")
    if ((camcol < 1) | (camcol > 6)).any():
        raise ValueError("camcol values are out-of-bounds!")
    if ((field < 0) | (field >= 2**12)).any():
        raise ValueError("camcol values are out-of-bounds!")
    if ((objnum < 0) | (objnum >= 2**16)).any():
        raise ValueError("id values are out-of-bounds!")
    #
    # Compute the objid
    #
    objid = ((skyversion << 59) |
             (rerun << 48) |
             (run << 32) |
             (camcol << 29) |
             (firstfield << 28) |
             (field << 16) |
             (objnum))
    return objid


def sdss_specobjid(plate, fiber, mjd, run2d, line=None, index=None):
    r"""Convert SDSS spectrum identifiers into CAS-style specObjID.

    Bits are assigned in specObjID thus:

    ===== ========== =============================================================
    Bits  Name       Comment
    ===== ========== =============================================================
    50-63 Plate ID   14 bits
    38-49 Fiber ID   12 bits
    24-37 MJD        Date plate was observed minus 50000 (14 bits)
    10-23 run2d      Spectroscopic reduction version
    0-9   line/index 0 for use in SpecObj files see below for other uses (10 bits)
    ===== ========== =============================================================

    Parameters
    ----------
    plate, fiber, mjd : :class:`int` or array of int
        Plate, fiber ID, and MJD for a spectrum.  If arrays are
        passed, all must have the same length.  The MJD value must be
        greater than 50000.
    run2d : :class:`int`, :class:`str` or array of int or str
        The run2d value must be an integer or a string of the form 'vN_M_P'.
        If an array is passed, it must have the same length as the other
        inputs listed above.  If the string form is used, the values are
        restricted to :math:`5 \le N \le 6`, :math:`0 \le M \le 99`,
        :math:`0 \le P \le 99`.
    line : :class:`int`, optional
        A line index, only used for defining specObjID for SpecLine files.
        `line` and `index` cannot both be non-zero.
    index : :class:`int`, optional
        An index measure, only used for defining specObjID for SpecLineIndex
        files. `line` and `index` cannot both be non-zero.

    Returns
    -------
    :class:`numpy.ndarray` of :class:`numpy.uint64`
        The specObjIDs of the objects.

    Raises
    ------
    :exc:`ValueError`
        If the sizes of the arrays don't match or if the array values are
        out of bounds.

    Notes
    -----
    * On 32-bit systems, makes sure to explicitly declare all inputs as
      64-bit integers.
    * This function defines the SDSS-III/IV version of specObjID, used for
      SDSS DR8 and subsequent data releases.  It is not compatible with
      SDSS DR7 or earlier.
    * If the string form of `run2d` is used, the bits are assigned by
      the formula :math:`(N - 5) \times 10000 + M \times 100 + P`.

    Examples
    --------
    >>> from pydl.pydlutils.sdss import sdss_specobjid
    >>> print(sdss_specobjid(4055,408,55359,'v5_7_0'))
    [4565636362342690816]
    """
    if line is not None and index is not None:
        raise ValueError("line and index inputs cannot both be non-zero!")
    if isinstance(plate, int):
        plate = np.array([plate])
    if isinstance(fiber, int):
        fiber = np.array([fiber])
    if isinstance(mjd, int):
        mjd = np.array([mjd]) - 50000
    if isinstance(run2d, str):
        try:
            run2d = np.array([int(run2d)])
        except ValueError:
            # Try a "vN_M_P" string.
            m = re.match(r'v(\d+)_(\d+)_(\d+)', run2d)
            if m is None:
                raise ValueError("Could not extract integer run2d value!")
            else:
                N, M, P = m.groups()
            run2d = np.array([(int(N) - 5)*10000 + int(M) * 100 + int(P)],
                             dtype=np.uint64)
    elif isinstance(run2d, int):
        run2d = np.array([run2d])
    if line is None:
        line = np.zeros(plate.shape, dtype=plate.dtype)
    else:
        if isinstance(line, int):
            line = np.array([line])
    if index is None:
        index = np.zeros(plate.shape, dtype=plate.dtype)
    else:
        if isinstance(index, int):
            index = np.array([index])
    #
    # Check that all inputs have the same shape.
    #
    if plate.shape != fiber.shape:
        raise ValueError("fiber.shape does not match plate.shape!")
    if plate.shape != mjd.shape:
        raise ValueError("mjd.shape does not match plate.shape!")
    if plate.shape != run2d.shape:
        raise ValueError("run2d.shape does not match plate.shape!")
    if plate.shape != line.shape:
        raise ValueError("line.shape does not match plate.shape!")
    if plate.shape != index.shape:
        raise ValueError("index.shape does not match plate.shape!")
    #
    # Check ranges of parameters
    #
    if ((plate < 0) | (plate >= 2**14)).any():
        raise ValueError("plate values are out-of-bounds!")
    if ((fiber < 0) | (fiber >= 2**12)).any():
        raise ValueError("fiber values are out-of-bounds!")
    if ((mjd < 0) | (mjd >= 2**14)).any():
        raise ValueError("MJD values are out-of-bounds!")
    if ((run2d < 0) | (run2d >= 2**14)).any():
        raise ValueError("MJD values are out-of-bounds!")
    if ((line < 0) | (line >= 2**10)).any():
        raise ValueError("line values are out-of-bounds!")
    if ((index < 0) | (index >= 2**10)).any():
        raise ValueError("index values are out-of-bounds!")
    #
    # Compute the specObjID
    #
    specObjID = ((plate.astype(np.uint64) << 50) |
                 (fiber.astype(np.uint64) << 38) |
                 (mjd.astype(np.uint64) << 24) |
                 (run2d.astype(np.uint64) << 10) |
                 (line.astype(np.uint64) | index.astype(np.uint64)))
    return specObjID


def sdss_sweep_circle(ra, dec, radius, stype='star', allobj=False):
    """Read the SDSS datasweep files and return objects around a location.

    Parameters
    ----------
    ra, dec : :class:`float`
        The sky location to search, J2000 degrees.
    radius : :class:`float`
        The radius around `ra`, `dec` to search.
    stype : :class:`str`, optional
        The type of object to search, 'star', 'gal' or 'sky'.
        The default is 'star'.
    allobj : :class:`bool`, optional
        If set to ``True``, return all objects found, not just SURVEY_PRIMARY.

    Returns
    -------
    :class:`numpy.ndarray`
        The data extracted from the sweep files.

    Raises
    ------
    :exc:`PydlutilsException`
        If :envvar:`PHOTO_SWEEP` is not set.

    Notes
    -----
    Assumes that the sweep files exist in :envvar:`PHOTO_SWEEP` and
    that index files have been created.
    """
    global sweep_cache
    #
    # Check values
    #
    if stype not in ('star', 'gal', 'sky'):
        raise ValueError('Invalid type {0}!'.format(stype))
    sweepdir = os.getenv('PHOTO_SWEEP')
    if sweepdir is None:
        raise PydlutilsException('PHOTO_SWEEP is not set!')
    #
    # Read the index
    #
    if sweep_cache[stype] is None:
        indexfile = os.path.join(sweepdir, 'datasweep-index-{0}.fits'.format(stype))
        with fits.open(indexfile) as f:
            sweep_cache[stype] = f[1].data
    index = sweep_cache[stype]
    #
    # Match
    #
    ira = np.array([ra])
    idec = np.array([dec])
    m1, m2, d12 = spherematch(index['RA'], index['DEC'], ira, idec,
                              radius+0.36, maxmatch=0)
    if len(m1) == 0:
        return None
    if not allobj:
        w = index['NPRIMARY'][m1] > 0
        if w.any():
            m1 = m1[w]
        else:
            return None
    #
    # Maximum number of objects
    #
    if allobj:
        n = index['IEND'][m1] - index['ISTART'][m1] + 1
        ntot = (np.where(n > 0, n, np.zeros(n.shape, dtype=n.dtype))).sum()
    else:
        ntot = index['NPRIMARY'][m1].sum()
    #
    # Find unique run + camcol
    #
    rc = index['RUN'][m1]*6 + index['CAMCOL'][m1] - 1
    isort = rc.argsort()
    iuniq = uniq(rc[isort])
    istart = 0
    objs = None
    nobjs = 0
    for i in range(len(iuniq)):
        iend = iuniq[i]
        icurr = isort[istart:iend]
        #
        # Determine which file and range of rows
        #
        run = index['RUN'][m1[icurr[0]]]
        camcol = index['CAMCOL'][m1[icurr[0]]]
        rerun = index['RERUN'][m1[icurr[0]]]
        fields = index[m1[icurr]]
        ist = fields['ISTART'].min()
        ind = fields['IEND'].max()
        if ind >= ist:
            #
            # Read in the rows of that file
            #
            swfile = os.path.join(os.getenv('PHOTO_SWEEP'), rerun,
                                  f'calibObj-{run:06d}-{camcol:1d}-{stype}.fits.gz')
            with fits.open(swfile) as f:
                tmp_objs = f[1].data[ist:ind]
            if tmp_objs.size > 0:
                #
                # Keep only objects within the desired radius
                #
                tm1, tm2, d12 = spherematch(tmp_objs['RA'], tmp_objs['DEC'],
                                            ira, idec, radius, maxmatch=0)
                if len(tm1) > 0:
                    tmp_objs = tmp_objs[tm1]
                    #
                    # Keep only SURVEY_PRIMARY objects by default
                    #
                    if not allobj:
                        w = ((tmp_objs['RESOLVE_STATUS'] &
                             sdss_flagval('RESOLVE_STATUS',
                                          'SURVEY_PRIMARY')) > 0)
                        if w.any():
                            tmp_objs = tmp_objs[w]
                        else:
                            tmp_objs = None
                    if tmp_objs is not None:
                        if objs is None:
                            objs = np.zeros(ntot, dtype=tmp_objs.dtype)
                        objs[nobjs:nobjs+tmp_objs.size] = tmp_objs
                        nobjs += tmp_objs.size
        istart = iend+1
    if nobjs > 0:
        return objs[0:nobjs]
    else:
        return None


def set_maskbits(idlutils_version='v5_5_33', maskbits_file=None):
    """Populate the maskbits cache.

    Parameters
    ----------
    idlutils_version : :class:`str`, optional
        Fetch the sdssMaskbits.par file corresponding to this idlutils version.
    maskbits_file : :class:`str`, optional
        Use an explicit file instead of downloading the official version.
        This should only be used for tests.

    Returns
    -------
    :class:`dict`
        A dictionary of bitmasks suitable for caching.

    Raises
    ------
    :exc:`URLError`
        If the data file could not be retrieved.
    """
    from astropy.utils.data import download_file
    from .yanny import yanny
    if maskbits_file is None:  # pragma: no cover
        if (idlutils_version == 'trunk' or
                idlutils_version.startswith('branches/')):
            iversion = idlutils_version
        else:
            iversion = 'tags/'+idlutils_version
        baseurl = ('https://svn.sdss.org/public/repo/sdss/idlutils/' +
                   '{0}/data/sdss/sdssMaskbits.par').format(iversion)
        filename = download_file(baseurl, cache=True, show_progress=False)
    else:
        filename = maskbits_file
    maskfile = yanny(filename, raw=True)
    #
    # Parse the file & cache the results in maskbits
    #
    maskbits = dict()
    for k in range(maskfile.size('MASKBITS')):
        if maskfile['MASKBITS']['flag'][k] in maskbits:
            maskbits[maskfile['MASKBITS']['flag'][k]][maskfile['MASKBITS']['label'][k]] = maskfile['MASKBITS']['bit'][k]
        else:
            maskbits[maskfile['MASKBITS']['flag'][k]] = {maskfile['MASKBITS']['label'][k]: maskfile['MASKBITS']['bit'][k]}
    if 'MASKALIAS' in maskfile:
        for k in range(maskfile.size('MASKALIAS')):
            maskbits[maskfile['MASKALIAS']['alias'][k]] = maskbits[maskfile['MASKALIAS']['flag'][k]].copy()
    return maskbits


def unwrap_specobjid(specObjID, run2d_integer=False, specLineIndex=False):
    """Unwrap CAS-style specObjID into plate, fiber, mjd, run2d.

    See :func:`~pydl.pydlutils.sdss.sdss_specobjid` for details on how the
    bits within a specObjID are assigned.

    Parameters
    ----------
    specObjID : :class:`numpy.ndarray`
        An array containing 64-bit integers or strings.  If strings are passed,
        they will be converted to integers internally.
    run2d_integer : :class:`bool`, optional
        If ``True``, do *not* attempt to convert the encoded run2d values
        to a string of the form 'vN_M_P'.
    specLineIndex : :class:`bool`, optional
        If ``True`` interpret any low-order bits as being an 'index'
        rather than a 'line'.

    Returns
    -------
    :class:`numpy.recarray`
        A record array with the same length as `specObjID`, with the columns
        'plate', 'fiber', 'mjd', 'run2d', 'line'.

    Examples
    --------
    >>> from numpy import array, uint64
    >>> from pydl.pydlutils.sdss import unwrap_specobjid
    >>> unwrap_specobjid(array([4565636362342690816], dtype=uint64))
    rec.array([(4055, 408, 55359, 'v5_7_0', 0)],
              dtype=[('plate', '<i4'), ('fiber', '<i4'), ('mjd', '<i4'), ('run2d', '<U8'), ('line', '<i4')])

    """
    if (specObjID.dtype.type is np.string_ or specObjID.dtype.type is np.unicode_):
        tempobjid = specObjID.astype(np.uint64)
    elif specObjID.dtype.type is np.uint64:
        tempobjid = specObjID.copy()
    else:
        raise ValueError('Unrecognized type for specObjID!')
    run2d_dtype = 'U8'
    if run2d_integer:
        run2d_dtype = 'i4'
    line = 'line'
    if specLineIndex:
        line = 'index'
    unwrap = np.recarray(specObjID.shape,
                         dtype=[('plate', 'i4'), ('fiber', 'i4'),
                                ('mjd', 'i4'), ('run2d', run2d_dtype),
                                (line, 'i4')])
    unwrap.plate = np.bitwise_and(tempobjid >> 50, 2**14 - 1)
    unwrap.fiber = np.bitwise_and(tempobjid >> 38, 2**12 - 1)
    unwrap.mjd = np.bitwise_and(tempobjid >> 24, 2**14 - 1) + 50000
    run2d = np.bitwise_and(tempobjid >> 10, 2**14 - 1)
    if run2d_integer:
        unwrap.run2d = run2d
    else:
        N = ((run2d // 10000) + 5).tolist()
        M = ((run2d % 10000) // 100).tolist()
        P = (run2d % 100).tolist()
        unwrap.run2d = ['v{0:d}_{1:d}_{2:d}'.format(n, m, p)
                        for n, m, p in zip(N, M, P)]
    unwrap[line] = np.bitwise_and(tempobjid, 2**10 - 1)
    return unwrap



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the top-level pydl functions.
"""
import pytest
import numpy as np
try:
    from astropy.tests.compat import assert_allclose
except ImportError:
    from numpy.testing import assert_allclose
from astropy.utils.data import get_pkg_data_filename
from ..file_lines import file_lines
from ..median import median
from ..pcomp import pcomp
from ..rebin import rebin
from ..smooth import smooth
from ..uniq import uniq


def test_file_lines():
    #
    # Find the test files
    #
    line_numbers = (1, 42, 137)
    plainfiles = [get_pkg_data_filename('t/this-file-contains-{0:d}-lines.txt'.format(l))
                  for l in line_numbers]
    gzfiles = [get_pkg_data_filename('t/this-file-contains-{0:d}-lines.txt.gz'.format(l))
               for l in line_numbers]
    for i, p in enumerate(plainfiles):
        n = file_lines(p)
        assert n == line_numbers[i]
    for i, p in enumerate(gzfiles):
        n = file_lines(p, compress=True)
        assert n == line_numbers[i]
    #
    # Test list passing
    #
    n = file_lines(plainfiles)
    assert tuple(n) == line_numbers
    n = file_lines(gzfiles, compress=True)
    assert tuple(n) == line_numbers
    #
    # Make sure empty files work
    #
    n = file_lines(get_pkg_data_filename('t/this-file-is-empty.txt'))
    assert n == 0


def test_median():
    odd_data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13],
                        dtype=np.float32)
    even_data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                         dtype=np.float32)
    assert median(odd_data) == 7
    assert median(odd_data, even=True) == 7
    assert median(even_data) == 7
    assert median(even_data, even=True) == 6.5
    assert (median(odd_data, 3) == odd_data).all()
    with pytest.raises(ValueError):
        _ = median(np.ones((9, 9, 9)), 3)
    odd_data2 = np.vstack((odd_data, odd_data, odd_data, odd_data, odd_data))
    assert (median(odd_data2, 3) == odd_data2).all()
    assert (median(odd_data2, axis=0) == odd_data).all()
    assert (median(odd_data2, axis=1) ==
            7*np.ones((odd_data2.shape[0],), dtype=odd_data2.dtype)).all()


def test_pcomp():
    test_data_file = get_pkg_data_filename('t/pcomp_data.txt')
    test_data = np.loadtxt(test_data_file, dtype='d', delimiter=',')
    with pytest.raises(ValueError):
        foo = pcomp(np.arange(10))
    pcomp_data = test_data[0:20, :]
    m = 4
    n = 20
    means = np.tile(pcomp_data.mean(0), n).reshape(pcomp_data.shape)
    newarray = pcomp_data - means
    foo = pcomp(newarray, covariance=True)
    #
    # This array is obtained from the IDL version of PCOMP.
    # It is only accurate up to an overall sign on each column.
    #
    derived = test_data[20:40, :]
    for k in range(m):
        assert_allclose(abs(foo.derived[:, k]), abs(derived[:, k]), 1e-4)
    coefficients = test_data[40:44, :]
    coefficientsT = coefficients.T
    for k in range(m):
        assert_allclose(abs(foo.coefficients[:, k]),
                        abs(coefficientsT[:, k]),
                        1e-4)
    eigenvalues = test_data[44, :]
    assert_allclose(foo.eigenvalues, eigenvalues, 1e-4)
    variance = test_data[45, :]
    assert_allclose(foo.variance, variance, 1e-4)
    #
    # Test the standardization.
    #
    foo = pcomp(pcomp_data, standardize=True, covariance=True)
    # for k in range(m):
    #     assert_allclose(abs(foo.derived[:, k]), abs(derived[:, k]), 1e-4)
    # for k in range(m):
    #     assert_allclose(abs(foo.coefficients[:, k]),
    #                     abs(coefficientsT[:, k]),
    #                     1e-4)
    eigenvalues = test_data[46, :]
    assert_allclose(foo.eigenvalues, eigenvalues, 1e-4)
    variance = test_data[47, :]
    assert_allclose(foo.variance, variance, 1e-4)
    # assert_allclose(foo.derived[0, :], np.array([-1.64153312,
    #                                              -9.12322038,
    #                                              1.41790708,
    #                                              -8.29359322]))
    #
    # Make sure correlation is working at least.
    #
    foo = pcomp(pcomp_data, standardize=True)
    assert_allclose(foo.eigenvalues, np.array([2.84968632e+00,
                                               1.00127640e+00,
                                               1.48380121e-01,
                                               6.57156222e-04]))
    assert_allclose(foo.variance, np.array([7.12421581e-01,
                                            2.50319100e-01,
                                            3.70950302e-02,
                                            1.64289056e-04]))


def test_rebin():
    x = np.arange(40)
    with pytest.raises(ValueError):
        r = rebin(x, d=(10, 10))
    with pytest.raises(ValueError):
        r = rebin(x, d=(70,))
    with pytest.raises(ValueError):
        r = rebin(x, d=(30,))
    x = np.array([[1.0, 2.0], [2.0, 3.0]])
    rexpect = np.array([[1.0, 2.0], [1.5, 2.5], [2.0, 3.0], [2.0, 3.0]])
    r = rebin(x, d=(4, 2))
    assert np.allclose(r, rexpect)
    rexpect = np.array([[1.0, 1.5, 2.0, 2.0], [2.0, 2.5, 3.0, 3.0]])
    r = rebin(x, d=(2, 4))
    assert np.allclose(r, rexpect)
    rexpect = np.array([[1.0, 2.0], [1.0, 2.0], [2.0, 3.0], [2.0, 3.0]])
    r = rebin(x, d=(4, 2), sample=True)
    assert np.allclose(r, rexpect)
    rexpect = np.array([[1.0, 1.0, 2.0, 2.0], [2.0, 2.0, 3.0, 3.0]])
    r = rebin(x, d=(2, 4), sample=True)
    assert np.allclose(r, rexpect)
    x = np.arange(10)
    rexpect = np.array([0.0, 2.0, 4.0, 6.0, 8.0])
    r = rebin(x, d=(5,), sample=True)
    assert np.allclose(r, rexpect)
    x = np.array([[1.0, 2.0, 3.0, 4.0],
                  [2.0, 3.0, 4.0, 5.0],
                  [3.0, 4.0, 5.0, 6.0],
                  [4.0, 5.0, 6.0, 7.0]])
    rexpect = np.array([[2.0, 4.0], [4.0, 6.0]])
    r = rebin(x, d=(2, 2))
    assert np.allclose(r, rexpect)
    rexpect = np.array([[1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 4.5],
                        [3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 6.5]])
    r = rebin(x, d=(2, 8))
    assert np.allclose(r, rexpect)


@pytest.mark.xfail
def test_rebin_int():
    """Test rebin on integers.  Comparing to IDL code similar to this::

        IDL> seed = 100
        IDL> array = FIX(RANDOMN(seed, 10, 20) * 100, TYPE=1) ; UINT8
        IDL> array_rebin = REBIN(array, 5, 10)
    """
    array = np.array([[188, 186, 25, 212, 34, 98, 3, 235, 155, 148],
                      [107, 166, 4, 41, 101, 190, 39, 154, 153, 239],
                      [135, 181, 92, 161, 213, 136, 35, 61, 80, 164],
                      [123, 248, 8, 157, 96, 118, 99, 1, 109, 246],
                      [226, 71, 183, 27, 46, 99, 8, 239, 66, 25],
                      [27, 219, 37, 130, 5, 81, 65, 250, 96, 14],
                      [71, 157, 156, 136, 47, 225, 247, 191, 49, 12],
                      [231, 133, 9, 38, 243, 2, 235, 145, 23, 22],
                      [146, 38, 49, 89, 42, 57, 220, 214, 135, 47],
                      [101, 116, 122, 209, 141, 37, 158, 224, 245, 82],
                      [15, 47, 51, 250, 207, 193, 209, 228, 110, 1],
                      [59, 232, 216, 224, 24, 118, 190, 10, 107, 27],
                      [84, 193, 112, 206, 113, 171, 138, 117, 244, 20],
                      [5, 31, 128, 214, 200, 119, 59, 27, 57, 10],
                      [226, 71, 177, 85, 0, 68, 54, 207, 141, 250],
                      [52, 119, 121, 177, 165, 99, 68, 29, 137, 200],
                      [172, 91, 181, 187, 87, 250, 45, 154, 58, 83],
                      [56, 175, 189, 35, 203, 223, 243, 187, 252, 97],
                      [186, 172, 207, 128, 61, 231, 89, 57, 131, 222],
                      [206, 96, 29, 60, 3, 8, 221, 55, 60, 17]], dtype=np.uint8)
    array_rebin = np.array([[161, 70, 105, 107, 173],
                            [171, 104, 140, 49, 149],
                            [135, 94, 57, 140, 50],
                            [148, 84, 129, 204, 26],
                            [100, 117, 69, 204, 127],
                            [88, 185, 135, 159, 61],
                            [78, 165, 150, 85, 82],
                            [116, 140, 83, 89, 181],
                            [123, 148, 190, 157, 122],
                            [165, 105, 75, 105, 107]], dtype=np.uint8)

    ar = rebin(array, (10, 5))
    assert (array_rebin == ar).all()


def test_rebin_int_sample():
    """Similar to test_rebin_int(), but using the sample option.
    """
    array = np.array([[188, 186, 25, 212, 34, 98, 3, 235, 155, 148],
                      [107, 166, 4, 41, 101, 190, 39, 154, 153, 239],
                      [135, 181, 92, 161, 213, 136, 35, 61, 80, 164],
                      [123, 248, 8, 157, 96, 118, 99, 1, 109, 246],
                      [226, 71, 183, 27, 46, 99, 8, 239, 66, 25],
                      [27, 219, 37, 130, 5, 81, 65, 250, 96, 14],
                      [71, 157, 156, 136, 47, 225, 247, 191, 49, 12],
                      [231, 133, 9, 38, 243, 2, 235, 145, 23, 22],
                      [146, 38, 49, 89, 42, 57, 220, 214, 135, 47],
                      [101, 116, 122, 209, 141, 37, 158, 224, 245, 82],
                      [15, 47, 51, 250, 207, 193, 209, 228, 110, 1],
                      [59, 232, 216, 224, 24, 118, 190, 10, 107, 27],
                      [84, 193, 112, 206, 113, 171, 138, 117, 244, 20],
                      [5, 31, 128, 214, 200, 119, 59, 27, 57, 10],
                      [226, 71, 177, 85, 0, 68, 54, 207, 141, 250],
                      [52, 119, 121, 177, 165, 99, 68, 29, 137, 200],
                      [172, 91, 181, 187, 87, 250, 45, 154, 58, 83],
                      [56, 175, 189, 35, 203, 223, 243, 187, 252, 97],
                      [186, 172, 207, 128, 61, 231, 89, 57, 131, 222],
                      [206, 96, 29, 60, 3, 8, 221, 55, 60, 17]], dtype=np.uint8)
    array_sample = np.array([[188, 25, 34, 3, 155],
                             [135, 92, 213, 35, 80],
                             [226, 183, 46, 8, 66],
                             [71, 156, 47, 247, 49],
                             [146, 49, 42, 220, 135],
                             [15, 51, 207, 209, 110],
                             [84, 112, 113, 138, 244],
                             [226, 177, 0, 54, 141],
                             [172, 181, 87, 45, 58],
                             [186, 207, 61, 89, 131]], dtype=np.uint8)
    ars = rebin(array, (10, 5), sample=True)
    assert (array_sample == ars).all()


def test_smooth():
    test_data_file = get_pkg_data_filename('t/smooth_data.txt')
    noise = np.loadtxt(test_data_file, dtype='d')
    #
    # Test smooth function
    #
    x = 8.0*np.arange(100)/100.0 - 4.0
    y = np.sin(x) + 0.1*noise
    s = smooth(y, 5)
    assert s.shape == (100,)
    s_edge = smooth(y, 5, True)
    assert s_edge.shape == (100,)
    s_w = smooth(y, 1)
    assert (s_w == y).all()


def test_uniq():
    items = np.array([1, 2, 3, 1, 5, 6, 1, 7, 3, 2, 5, 9, 11, 1])
    items_sorted = np.sort(items)
    #
    # On some systems, kind='quicksort', which is the default, gives different
    # results depending on numpy version.
    #
    items_argsorted = np.argsort(items, kind='stable')
    #
    # Test pre-sorted array.
    #
    u1 = uniq(items_sorted)
    assert (u1 == np.array([3, 5, 7, 9, 10, 11, 12, 13])).all()
    #
    # Test arg-sorted array.
    #
    u2 = uniq(items, items_argsorted)
    assert (u2 == np.array([13, 9, 8, 10, 5, 7, 11, 12])).all()
    assert (items_sorted[u1] == items[u2]).all()
    #
    # Test degenerate case of all identical items.
    #
    identical_items = np.ones((10,), dtype=items.dtype)
    u = uniq(identical_items)
    assert (u == np.array([9])).all()
    u = uniq(identical_items, np.arange(10, dtype=items.dtype))
    assert (u == np.array([9])).all()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the photoobj directory in photoop.
"""
import numpy as np
import astropy.units as u


def sdss_calibv():
    """Return calibration for velocities from pix/frame to deg/day.

    Returns
    -------
    :class:`~astropy.units.quanity.Quantity`
        The conversion from pixels per frame to degrees per day

    Notes
    -----
    Assumes frame time difference of 71.72 seconds and pixel scale of
    0.396 arcsec, both fixed. Also note that observations of the same part of
    sky from adjacent bands are separated by *two* frame numbers,
    so we multiply by a factor two.
    """
    pixscale = 0.396 * u.arcsec
    ftime = 71.72 * u.s
    return (2.0*pixscale/ftime).to('deg / d')


def unwrap_objid(objid):
    """Unwrap CAS-style objID into run, camcol, field, id, rerun.

    See :func:`~pydl.pydlutils.sdss.sdss_objid` for details on how the bits
    within an objID are assigned.

    Parameters
    ----------
    objid : :class:`numpy.ndarray`
        An array containing 64-bit integers or strings.  If strings are passed,
        they will be converted to integers internally.

    Returns
    -------
    :class:`numpy.recarray`
        A record array with the same length as `objid`, with the columns
        'skyversion', 'rerun', 'run', 'camcol', 'firstfield', 'frame', 'id'.

    Raises
    ------
    :exc:`ValueError`
        If the input objID has a type that can't be converted into 64-bit integer.

    Notes
    -----
    For historical reasons, the inverse of this function,
    :func:`~pydl.pydlutils.sdss.sdss_objid` is not in the same namespace as
    this function.

    'frame' is used instead of 'field' because record arrays have a method
    of the same name.

    Examples
    --------
    >>> from numpy import array
    >>> from pydl.photoop.photoobj import unwrap_objid
    >>> unwrap_objid(array([1237661382772195474]))
    rec.array([(2, 301, 3704, 3, 0, 91, 146)],
          dtype=[('skyversion', '<i4'), ('rerun', '<i4'), ('run', '<i4'), ('camcol', '<i4'), ('firstfield', '<i4'), ('frame', '<i4'), ('id', '<i4')])
    """
    if objid.dtype.type is np.string_ or objid.dtype.type is np.unicode_:
        tempobjid = objid.astype(np.int64)
    elif objid.dtype.type is np.int64:
        tempobjid = objid.copy()
    else:
        raise ValueError('Unrecognized type for objid!')
    unwrap = np.recarray(objid.shape,
                         dtype=[('skyversion', 'i4'), ('rerun', 'i4'),
                                ('run', 'i4'), ('camcol', 'i4'),
                                ('firstfield', 'i4'),
                                ('frame', 'i4'), ('id', 'i4')])
    unwrap.skyversion = np.bitwise_and(tempobjid >> 59, 2**4 - 1)
    unwrap.rerun = np.bitwise_and(tempobjid >> 48, 2**11 - 1)
    unwrap.run = np.bitwise_and(tempobjid >> 32, 2**16 - 1)
    unwrap.camcol = np.bitwise_and(tempobjid >> 29, 2**3 - 1)
    unwrap.firstfield = np.bitwise_and(tempobjid >> 28, 2**1 - 1)
    unwrap.frame = np.bitwise_and(tempobjid >> 16, 2**12 - 1)
    unwrap.id = np.bitwise_and(tempobjid, 2**16 - 1)
    return unwrap



"""Configure Test Suite.

This file is used to configure the behavior of pytest when using the Astropy
test infrastructure. It needs to live inside the package in order for it to
get picked up when running the tests inside an interpreter using
``pydl.test()``

"""

# import os

# from astropy.version import version as astropy_version

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


def pytest_configure(config):
    """Configure Pytest with Astropy.

    Parameters
    ----------
    config : pytest configuration

    """
    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES.pop('h5py', None)
        PYTEST_HEADER_MODULES['PyDL'] = 'pydl'

        from . import __version__
        TESTED_VERSIONS['pydl'] = __version__



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the trace directory in idlutils.

Traceset (or "trace set"):
    A set of coefficient vectors defining a set of functions
    over a common independent-variable domain specified by ``xmin``
    and ``xmax`` values. The functions in the set are defined in
    terms of a linear combination of basis functions (such as
    Legendre of Chebyshev polynonials) up to a specified maximum
    order, weighted by the values in the coefficient vectors, and
    evaluated using a suitable affine rescaling of the
    1dependent-variable domain (such as ``[xmin, xmax] -> [-1, 1]``
    for Legendre polynomials). For evaluation purposes, the
    domain is assumed by default to be a zero-based integer
    baseline from ``xmin`` to ``xmax`` such as would correspond to a
    digital detector pixel grid.

Etymology:
    From the original motivating use case of encoding the "traces"
    of multiple spectra across the detector in a multi-fiber spectrograph.

Definition from A. Bolton, U. of Utah, June 2015.
"""
import numpy as np
from scipy.special import chebyt
from astropy.io.fits.fitsrec import FITS_rec
from . import PydlutilsException
from .math import djs_reject
from .misc import djs_laxisgen
from ..goddard.math import flegendre


def fchebyshev(x, m):
    """Compute the first `m` Chebyshev polynomials.

    Parameters
    ----------
    x : array-like
        Compute the Chebyshev polynomials at these abscissa values.
    m : :class:`int`
        The number of Chebyshev polynomials to compute.  For example, if
        :math:`m = 3`, :math:`T_0 (x)`, :math:`T_1 (x)` and
        :math:`T_2 (x)` will be computed.

    Returns
    -------
    :class:`numpy.ndarray`
    """
    if isinstance(x, np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 1:
        raise ValueError('Order of Chebyshev polynomial must be at least 1.')
    try:
        dt = x.dtype
    except AttributeError:
        dt = np.float64
    leg = np.ones((m, n), dtype=dt)
    if m >= 2:
        leg[1, :] = x
    if m >= 3:
        for k in range(2, m):
            leg[k, :] = np.polyval(chebyt(k), x)
    return leg


def fchebyshev_split(x, m):
    """Compute the first `m` Chebyshev polynomials, but modified to allow a
    split in the baseline at :math:`x=0`.  The intent is to allow a model fit
    where a constant term is different for positive and negative `x`.

    Parameters
    ----------
    x : array-like
        Compute the Chebyshev polynomials at these abscissa values.
    m : :class:`int`
        The number of Chebyshev polynomials to compute.  For example, if
        :math:`m = 3`, :math:`T_0 (x)`, :math:`T_1 (x)` and
        :math:`T_2 (x)` will be computed.

    Returns
    -------
    :class:`numpy.ndarray`
    """
    if isinstance(x, np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 2:
        raise ValueError('Order of polynomial must be at least 2.')
    try:
        dt = x.dtype
    except AttributeError:
        dt = np.float64
    leg = np.ones((m, n), dtype=dt)
    try:
        leg[0, :] = (x >= 0).astype(x.dtype)
    except AttributeError:
        leg[0, :] = np.double(x >= 0)
    if m > 2:
        leg[2, :] = x
    if m > 3:
        for k in range(3, m):
            leg[k, :] = 2.0 * x * leg[k-1, :] - leg[k-2, :]
    return leg


def fpoly(x, m):
    """Compute the first `m` simple polynomials.

    Parameters
    ----------
    x : array-like
        Compute the simple polynomials at these abscissa values.
    m : :class:`int`
        The number of simple polynomials to compute.  For example, if
        :math:`m = 3`, :math:`x^0`, :math:`x^1` and
        :math:`x^2` will be computed.

    Returns
    -------
    :class:`numpy.ndarray`
    """
    if isinstance(x, np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 1:
        raise ValueError('Order of polynomial must be at least 1.')
    try:
        dt = x.dtype
    except AttributeError:
        dt = np.float64
    leg = np.ones((m, n), dtype=dt)
    if m >= 2:
        leg[1, :] = x
    if m >= 3:
        for k in range(2, m):
            leg[k, :] = leg[k-1, :] * x
    return leg


def func_fit(x, y, ncoeff, invvar=None, function_name='legendre', ia=None,
             inputans=None, inputfunc=None):
    """Fit `x`, `y` positions to a functional form.

    Parameters
    ----------
    x : array-like
        X values (independent variable).
    y : array-like
        Y values (dependent variable).
    ncoeff : :class:`int`
        Number of coefficients to fit.
    invvar : array-like, optional
        Weight values; inverse variance.
    function_name : :class:`str`, optional
        Function name, default 'legendre'.
    ia : array-like, optional
        An array of bool of length `ncoeff` specifying free (``True``) and
        fixed (``False``) parameters.
    inputans : array-like, optional
        An array of values of length `ncoeff` specifying the values of
        the fixed parameters.
    inputfunc : array-like, optional
        Multiply the function fit by these values.

    Returns
    -------
    :class:`tuple` of array-like
        Fit coefficients, length `ncoeff`; fitted values.

    Raises
    ------
    :exc:`KeyError`
        If an invalid function type is selected.
    :exc:`ValueError`
        If input dimensions do not agree.
    """
    if x.shape != y.shape:
        raise ValueError('Dimensions of X and Y do not agree!')
    if invvar is None:
        invvar = np.ones(x.shape, dtype=x.dtype)
    else:
        if invvar.shape != x.shape:
            raise ValueError('Dimensions of X and invvar do not agree!')
    if ia is None:
        ia = np.ones((ncoeff,), dtype=bool)
    if not ia.all():
        if inputans is None:
            inputans = np.zeros((ncoeff,), dtype=x.dtype)
    #
    # Select unmasked points
    #
    igood = (invvar > 0).nonzero()[0]
    ngood = len(igood)
    res = np.zeros((ncoeff,), dtype=x.dtype)
    yfit = np.zeros(x.shape, dtype=x.dtype)
    if ngood == 0:
        pass
    elif ngood == 1:
        res[0] = y[igood[0]]
        yfit += y[igood[0]]
    else:
        ncfit = min(ngood, ncoeff)
        function_map = {'legendre': flegendre,
                        'flegendre': flegendre,
                        'chebyshev': fchebyshev,
                        'fchebyshev': fchebyshev,
                        'chebyshev_split': fchebyshev_split,
                        'fchebyshev_split': fchebyshev_split,
                        'poly': fpoly,
                        'fpoly': fpoly}
        try:
            legarr = function_map[function_name](x, ncfit)
        except KeyError:
            raise KeyError('Unknown function type: {0}'.format(function_name))
        if inputfunc is not None:
            if inputfunc.shape != x.shape:
                raise ValueError('Dimensions of X and inputfunc do not agree!')
            legarr *= np.tile(inputfunc, ncfit).reshape(ncfit, x.shape[0])
        yfix = np.zeros(x.shape, dtype=x.dtype)
        nonfix = ia[0:ncfit].nonzero()[0]
        nparams = len(nonfix)
        fixed = (~ia[0:ncfit]).nonzero()[0]
        if len(fixed) > 0:
            yfix = np.dot(legarr.T, inputans * (1 - ia))
            ysub = y - yfix
            finalarr = legarr[nonfix, :]
        else:
            finalarr = legarr
            ysub = y
        # extra2 = finalarr * np.outer(np.ones((nparams,), dtype=x.dtype),
        #                             (invvar > 0))
        extra2 = finalarr * np.outer(np.ones((nparams,), dtype=x.dtype),
                                     invvar)
        alpha = np.dot(finalarr, extra2.T)
        # assert alpha.dtype == x.dtype
        if nparams > 1:
            # beta = np.dot(ysub * (invvar > 0), finalarr.T)
            beta = np.dot(ysub * invvar, finalarr.T)
            assert beta.dtype == x.dtype
            # uu,ww,vv = np.linalg.svd(alpha, full_matrices=False)
            res[nonfix] = np.linalg.solve(alpha, beta)
        else:
            # res[nonfix] = (ysub * (invvar > 0) * finalarr).sum()/alpha
            res[nonfix] = (ysub * invvar * finalarr).sum()/alpha
        if len(fixed) > 0:
            res[fixed] = inputans[fixed]
        yfit = np.dot(legarr.T, res[0:ncfit])
    return (res, yfit)


class TraceSet(object):
    """Implements the idea of a trace set.

    Attributes
    ----------
    func : :class:`str`
        Name of function type used to fit the trace set.
    xmin : float-like
        Minimum x value.
    xmax : float-like
        Maximum x value.
    coeff : array-like
        Coefficients of the trace set fit.
    nTrace : :class:`int`
        Number of traces in the object.
    ncoeff : :class:`int`
        Number of coefficients of the trace set fit.
    xjumplo : float-like
        Jump value, for BOSS readouts.
    xjumphi : float-like
        Jump value, for BOSS readouts.
    xjumpval : float-like
        Jump value, for BOSS readouts.
    outmask : array-like
        When initialized with x,y positions, this contains the rejected
        points.
    yfit : array-like
        When initialized with x,y positions, this contains the fitted y
        values.
    """
    _func_map = {'poly': fpoly, 'legendre': flegendre, 'chebyshev': fchebyshev}

    def __init__(self, *args, **kwargs):
        """This class can be initialized either with a set of xy positions,
        or with a trace set HDU from a FITS file.
        """
        if len(args) == 1 and isinstance(args[0], FITS_rec):
            #
            # Initialize with FITS data
            #
            self.func = args[0]['FUNC'][0]
            self.xmin = args[0]['XMIN'][0]
            self.xmax = args[0]['XMAX'][0]
            self.coeff = args[0]['COEFF'][0]
            self.nTrace = self.coeff.shape[0]
            self.ncoeff = self.coeff.shape[1]
            if 'XJUMPLO' in args[0].dtype.names:
                self.xjumplo = args[0]['XJUMPLO'][0]
                self.xjumphi = args[0]['XJUMPHI'][0]
                self.xjumpval = args[0]['XJUMPVAL'][0]
            else:
                self.xjumplo = None
                self.xjumphi = None
                self.xjumpval = None
            self.outmask = None
            self.yfit = None
        elif len(args) == 2:
            #
            # Initialize with x, y positions.
            #
            xpos = args[0]
            ypos = args[1]
            self.nTrace = xpos.shape[0]
            if 'invvar' in kwargs:
                invvar = kwargs['invvar']
            else:
                invvar = np.ones(xpos.shape, dtype=xpos.dtype)
            if 'func' in kwargs:
                self.func = kwargs['func']
            else:
                self.func = 'legendre'
            if 'ncoeff' in kwargs:
                self.ncoeff = int(kwargs['ncoeff'])
            else:
                self.ncoeff = 3
            if 'xmin' in kwargs:
                self.xmin = np.float64(kwargs['xmin'])
            else:
                self.xmin = xpos.min()
            if 'xmax' in kwargs:
                self.xmax = np.float64(kwargs['xmax'])
            else:
                self.xmax = xpos.max()
            if 'maxiter' in kwargs:
                maxiter = int(kwargs['maxiter'])
            else:
                maxiter = 10
            if 'inmask' in kwargs:
                inmask = kwargs['inmask']
            else:
                inmask = np.ones(xpos.shape, dtype=bool)
            do_jump = False
            if 'xjumplo' in kwargs:
                do_jump = True
                self.xjumplo = np.float64(kwargs['xjumplo'])
            else:
                self.xjumplo = None
            if 'xjumphi' in kwargs:
                self.xjumphi = np.float64(kwargs['xjumphi'])
            else:
                self.xjumphi = None
            if 'xjumpval' in kwargs:
                self.xjumpval = np.float64(kwargs['xjumpval'])
            else:
                self.xjumpval = None
            self.coeff = np.zeros((self.nTrace, self.ncoeff), dtype=xpos.dtype)
            self.outmask = np.zeros(xpos.shape, dtype=bool)
            self.yfit = np.zeros(xpos.shape, dtype=xpos.dtype)
            for iTrace in range(self.nTrace):
                xvec = self.xnorm(xpos[iTrace, :], do_jump)
                iIter = 0
                qdone = False
                tempivar = (invvar[iTrace, :] *
                            inmask[iTrace, :].astype(invvar.dtype))
                thismask = tempivar > 0
                while (not qdone) and (iIter <= maxiter):
                    res, ycurfit = func_fit(xvec, ypos[iTrace, :], self.ncoeff,
                                            invvar=tempivar, function_name=self.func)
                    thismask, qdone = djs_reject(ypos[iTrace, :], ycurfit,
                                                 invvar=tempivar)
                    iIter += 1
                self.yfit[iTrace, :] = ycurfit
                self.coeff[iTrace, :] = res
                self.outmask[iTrace, :] = thismask
        else:
            raise PydlutilsException("Wrong number of arguments to TraceSet!")

    def xy(self, xpos=None, ignore_jump=False):
        """Convert from a trace set to an array of x,y positions.

        Parameters
        ----------
        xpos : array-like, optional
            If provided, evaluate the trace set at these positions.  Otherwise
            the positions will be constructed from the trace set object iself.
        ignore_jump : :class:`bool`, optional
            If ``True``, ignore any jump information in the `tset` object

        Returns
        -------
        :class:`tuple` of array-like
            The x, y positions.
        """
        do_jump = self.has_jump and (not ignore_jump)
        if xpos is None:
            xpos = djs_laxisgen([self.nTrace, self.nx], iaxis=1) + self.xmin
        ypos = np.zeros(xpos.shape, dtype=xpos.dtype)
        for iTrace in range(self.nTrace):
            xvec = self.xnorm(xpos[iTrace, :], do_jump)
            legarr = self._func_map[self.func](xvec, self.ncoeff)
            ypos[iTrace, :] = np.dot(legarr.T, self.coeff[iTrace, :])
        return (xpos, ypos)

    @property
    def has_jump(self):
        """``True`` if jump conditions are set.
        """
        return self.xjumplo is not None

    @property
    def xRange(self):
        """Range of x values.
        """
        return self.xmax - self.xmin

    @property
    def nx(self):
        """Number of x values.
        """
        return int(self.xRange + 1)

    @property
    def xmid(self):
        """Midpoint of x values.
        """
        return 0.5 * (self.xmin + self.xmax)

    def xnorm(self, xinput, jump):
        """Convert input x coordinates to normalized coordinates suitable
        for input to special polynomials.

        Parameters
        ----------
        xinput : array-like
            Input coordinates.
        jump : :class:`bool`
            Set to ``True`` if there is a jump.

        Returns
        -------
        array-like
            Normalized coordinates.
        """
        if jump:
            # Vector specifying what fraction of the jump has passed:
            jfrac = np.minimum(np.maximum(((xinput - self.xjumplo) /
                                           (self.xjumphi - self.xjumplo)), 0.), 1.)
            # Conversion to "natural" x baseline:
            xnatural = xinput + jfrac * self.xjumpval
        else:
            xnatural = xinput
        return 2.0 * (xnatural - self.xmid)/self.xRange


def traceset2xy(tset, xpos=None, ignore_jump=False):
    """Convert from a trace set to an array of x,y positions.

    Parameters
    ----------
    tset : :class:`TraceSet`
        A :class:`TraceSet` object.
    xpos : array-like, optional
        If provided, evaluate the trace set at these positions.  Otherwise
        the positions will be constructed from the trace set object iself.
    ignore_jump : bool, optional
        If ``True``, ignore any jump information in the `tset` object

    Returns
    -------
    :class:`tuple` of array-like
        The x, y positions.
    """
    return tset.xy(xpos, ignore_jump)


def xy2traceset(xpos, ypos, **kwargs):
    """Convert from x,y positions to a trace set.

    Parameters
    ----------
    xpos, ypos : array-like
        X,Y positions corresponding as [nx,Ntrace] arrays.
    invvar : array-like, optional
        Inverse variances for fitting.
    func : :class:`str`, optional
        Function type for fitting; defaults to 'legendre'.
    ncoeff : :class:`int`, optional
        Number of coefficients to fit.  Defaults to 3.
    xmin, xmax : :class:`float`, optional
        Explicitly set minimum and maximum values, instead of computing
        them from `xpos`.
    maxiter : :class:`int`, optional
        Maximum number of rejection iterations; set to 0 for no rejection;
        default to 10.
    inmask : array-like, optional
        Mask set to 1 for good points and 0 for rejected points;
        same dimensions as `xpos`, `ypos`.  Points rejected by `inmask`
        are always rejected from the fits (the rejection is "sticky"),
        and will also be marked as rejected in the outmask attribute.
    ia, inputans, inputfunc : array-like, optional
        These arguments will be passed to :func:`func_fit`.
    xjumplo : :class:`float`, optional
        x position locating start of an x discontinuity
    xjumphi : :class:`float`, optional
        x position locating end of that x discontinuity
    xjumpval : :class:`float`, optional
        magnitude of the discontinuity "jump" between those bounds
        (previous 3 keywords motivated by BOSS 2-phase readout)

    Returns
    -------
    :class:`TraceSet`
        A :class:`TraceSet` object.
    """
    return TraceSet(xpos, ypos, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlspec2d.spec2d.
"""
import os
import pytest
import numpy as np
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from ..spec2d import aesthetics, filter_thru
from .. import Pydlspec2dException


@pytest.fixture
def sdss_env(request):
    """Set up spectroscopic pipeline environment variables.
    """
    m = request.getfixturevalue("monkeypatch")
    e = {'BOSS_SPECTRO_REDUX': '/boss/spectro/redux',
         'SPECTRO_REDUX': '/sdss/spectro/redux',
         'RUN2D': 'v1_2_3',
         'RUN1D': 'v1_2_3'}
    for k in e:
        m.setenv(k, e[k])
    return m


def test_aesthetics():
    np.random.seed(137)
    flux = np.random.rand(100)
    ivar = np.random.rand(100)
    #
    # No bad
    #
    f = aesthetics(flux, ivar)
    assert (f == flux).all()
    #
    # Bad points
    #
    ivar[ivar < 0.1] = 0.0
    #
    # Bad method
    #
    with pytest.raises(Pydlspec2dException):
        f = aesthetics(flux, ivar, 'badmethod')
    #
    # Nothing
    #
    f = aesthetics(flux, ivar, 'nothing')
    assert (f == flux).all()


def test_combine1fiber():
    pass


def test_filter_thru(sdss_env):
    fname = get_pkg_data_filename('t/spPlate-4055-55359-0020.fits')
    with fits.open(fname) as hdulist:
        flux = hdulist[0].data
        npix = hdulist[0].header['NAXIS1']
        ntrace = hdulist[0].header['NAXIS2']
        crval1 = hdulist[0].header['COEFF0']
        cd1_1 = hdulist[0].header['COEFF1']
    assert flux.shape == (ntrace, npix)
    loglam0 = crval1 + cd1_1*np.arange(npix, dtype=flux.dtype)
    waveimg = 10**(np.tile(loglam0, 20).reshape(flux.shape))
    assert waveimg.shape == flux.shape
    f = filter_thru(flux, waveimg=waveimg)
    idl_data_file = get_pkg_data_filename('t/filter_thru_idl_data.txt')
    idl_data = np.loadtxt(idl_data_file, dtype='f', delimiter=',').T
    assert f.shape == (20, 5)
    assert np.allclose(f, idl_data, atol=1.0e-6)
    #
    # Test bad input.
    #
    with pytest.raises(ValueError):
        f = filter_thru(flux)
    with pytest.raises(ValueError):
        f = filter_thru(flux, waveimg=waveimg, filter_prefix='sdss')
    return


def prepare_data():
    """Convert full spPlate file into a test-sized version.
    """
    nTrace = 20
    spPlate = os.path.join(os.getenv('HOME'), 'Downloads',
                           'spPlate-4055-55359.fits')
    spPlateOut = os.path.join(os.path.dirname(__file__), 't',
                              'spPlate-4055-55359-0020.fits')
    with fits.open(spPlate) as hdulist:
        newhdu = fits.PrimaryHDU(hdulist[0].data[0:nTrace, :],
                                 header=hdulist[0].header)
        newhdulist = fits.HDUList([newhdu])
        newhdulist.writeto(spPlateOut)
    return 0


if __name__ == '__main__':
    from sys import exit
    exit(prepare_data())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the mangle directory in idlutils.

Mangle [1]_ is a software suite that supports the concept of polygons on
a sphere, and is used to, for example, describe the `window function`_ of the
Sloan Digital Sky Survey.

This implementation is intended to support the portions of Mangle that
are included in idlutils.  To simplify the coding somewhat, unlike
idlutils, the caps information is accessed through ``polygon.x`` and
``polygon.cm``, not ``polygon.caps.x`` or ``polygon.caps.cm``.

Window function operations are already supported by :func:`~is_in_polygon` and
:func:`~is_in_window`.  However, calculation of polygon area (solid angle)
from first principles is not yet implemented.

Note that in traditional geometry "spherical polygon" means a figure
bounded by *great circles*.  Mangle allows polygons to be bounded by
*any* circle, great or not.  Use care when comparing formulas in this
module to formulas in the mathematical literature.

.. _`window function`: https://www.sdss4.org/dr16/algorithms/resolve/

References
----------

.. [1] `Swanson, M. E. C.; Tegmark, Max; Hamilton, Andrew J. S.;
   Hill, J. Colin, 2008 MNRAS 387, 1391
   <https://ui.adsabs.harvard.edu/abs/2008MNRAS.387.1391S/abstract>`_.
"""
import re
import warnings
import numpy as np
from astropy.io import fits
# import astropy.utils as au
from . import PydlutilsException, PydlutilsUserWarning


class PolygonList(list):
    """A :class:`list` that contains :class:`ManglePolygon` objects and
    possibly some metadata.

    Parameters
    ----------
    header : :class:`list`, optional
        Set the `header` attribute.

    Attributes
    ----------
    header : :class:`list`
        A list of strings containing metadata.
    """
    def __init__(self, *args, **kwargs):
        super(PolygonList, self).__init__(*args)
        if 'header' in kwargs:
            self.header = kwargs['header']
        else:
            self.header = list()
        return


class ManglePolygon(object):
    """Simple object to represent a polygon.

    A polygon may be instantiated with a row
    (:class:`~astropy.io.fits.fitsrec.FITS_record`) from a
    :class:`FITS_polygon` object, another :class:`ManglePolygon` object
    (copy constructor), keyword arguments, or with no arguments at all,
    in which case it represents the whole sky.

    Attributes
    ----------
    cm : :class:`~numpy.ndarray`
        The size of each cap in the polygon.
    id : :class:`int`
        An arbitrary ID number.
    pixel : :class:`int`
        Pixel this polygon is in.
    use_caps : :class:`int`
        Bitmask indicating which caps to use.
    weight : :class:`float`
        Weight factor assigned to the polygon.
    """

    def __init__(self, *args, **kwargs):
        try:
            a0 = args[0]
        except IndexError:
            a0 = None
        if isinstance(a0, fits.fitsrec.FITS_record):
            self._ncaps = int(a0['NCAPS'])
            self.weight = float(a0['WEIGHT'])
            self.pixel = int(a0['PIXEL'])
            try:
                self.id = int(a0['IFIELD'])
            except KeyError:
                self.id = -1
            self._str = float(a0['STR'])
            self.use_caps = int(a0['USE_CAPS'])
            xcaps = a0['XCAPS']
            if xcaps.shape == (3, ):
                self._x = np.zeros((1, 3), dtype=xcaps.dtype) + xcaps
            else:
                self._x = xcaps[0:self._ncaps, :].copy()
            # assert self._x.shape == (self._ncaps, 3)
            cmcaps = a0['CMCAPS']
            if cmcaps.shape == ():
                self.cm = np.zeros((1,), dtype=cmcaps.dtype) + cmcaps
            else:
                self.cm = cmcaps[0:self._ncaps].copy()
            # assert self.cm.shape == (self._ncaps, )
        elif isinstance(a0, ManglePolygon):
            self._ncaps = a0._ncaps
            self.weight = a0.weight
            self.pixel = a0.pixel
            self.id = a0.id
            self._str = a0._str
            self.use_caps = a0.use_caps
            self._x = a0._x.copy()
            self.cm = a0.cm.copy()
        elif kwargs:
            if 'x' in kwargs and 'cm' in kwargs:
                xs = kwargs['x'].shape
                cm = kwargs['cm'].shape
                assert xs[0] == cm[0]
            else:
                raise ValueError('Input values are missing!')
            self._ncaps = xs[0]
            if 'weight' in kwargs:
                self.weight = float(kwargs['weight'])
            else:
                self.weight = 1.0
            if 'pixel' in kwargs:
                self.pixel = int(kwargs['pixel'])
            else:
                self.pixel = -1
            if 'id' in kwargs:
                self.id = int(kwargs['id'])
            else:
                self.id = -1
            if 'use_caps' in kwargs:
                self.use_caps = int(kwargs['use_caps'])
            else:
                # Use all caps by default
                self.use_caps = (1 << self.ncaps) - 1
            self._x = kwargs['x'].copy()
            self.cm = kwargs['cm'].copy()
            if 'str' in kwargs:
                self._str = float(kwargs['str'])
            else:
                self._str = None
        else:
            #
            # An "empty" polygon represents the whole sky.
            #
            self._ncaps = 0
            self.weight = 1.0
            self.pixel = -1
            self.id = -1
            self.use_caps = 0
            self._x = None
            self.cm = None
            self._str = 4.0*np.pi
        return

    @property
    def ncaps(self):
        """Number of caps in the polygon.
        """
        return self._ncaps

    @property
    def x(self):
        """The orientation of each cap in the polygon.  The direction is
        specified by a unit 3-vector.
        """
        return self._x

    @property
    def str(self):
        """Solid angle of this polygon (steradians).
        """
        if self._str is None:
            self._str = self.garea()
        return self._str

    def cmminf(self):
        """The index of the smallest cap in the polygon, accounting for
        negative caps and ``use_caps``.

        Returns
        -------
        :class:`int`
            Integer index of the smallest cap.
        """
        if self.ncaps == 0:
            return None
        cmmin = 2.0
        kmin = -1
        for k in range(self.ncaps):
            if is_cap_used(self.use_caps, k):
                if self.cm[k] >= 0:
                    cmk = self.cm[k]
                else:
                    cmk = 2.0 + self.cm[k]
                if cmk < cmmin:
                    cmmin = cmk
                    kmin = k
        return kmin

    def garea(self):
        r"""Compute the area of a polygon.

        See [1]_ for the detailed area formula, which is summarized here:

        * An empty polygon with no caps is defined to be the whole sky.
        * A polygon with one cap has area ``2*pi*self.cm``.
        * A polygon with at least one cap with an area less than :math:`2\pi`
          has an area less than :math:`2\pi`.
        * If every cap has an area greater than :math:`2\pi`, split the polygon
          into two smaller polygons and sum the two areas.

        Returns
        -------
        :class:`float`
            The area of the polygon.

        References
        ----------

        .. [1] `Hamilton, A. J. S.; Tegmark, Max, 2004 MNRAS 349, 115
           <https://ui.adsabs.harvard.edu/abs/2004MNRAS.349..115H/abstract>`_.
        """
        smallest_cap = self.cmminf()
        if smallest_cap is None:
            return 4.0 * np.pi
        cmmin = self.cm[smallest_cap]
        used_caps = list()
        for k in range(self.ncaps):
            if is_cap_used(self.use_caps, k):
                used_caps.append(k)
        nused_caps = len(used_caps)
        if nused_caps == 0:
            return 0.0
        if nused_caps >= 2 and cmmin > 1.0:
            npl = nused_caps + 1
        else:
            npl = nused_caps
        if npl == nused_caps:
            if nused_caps == 1:
                #
                # Short-circuit this case.
                #
                if cmmin < 0:
                    return 2.0*np.pi*(2.0+cmmin)
                else:
                    return 2.0*np.pi*cmmin
            else:
                #
                # Two or more caps, at least one has area < 2.0*pi.
                #
                return self._garea_helper()
        else:
            #
            # More than two caps, and all have area > 2.0*pi.
            # This section still needs to account for use_caps.
            #
            dpoly = self.polyn(self, smallest_cap)
            dpoly.cm[nused_caps] = cmmin / 2.0
            area1 = dpoly._garea_helper()
            dpoly.cm[nused_caps] = -1.0 * dpoly.cm[nused_caps]
            area2 = dpoly._garea_helper()
            return area1 + area2

    def _garea_helper(self):
        """Reproduces the Fortran routine garea in Mangle.

        *Placeholder for now.*

        Returns
        -------
        :class:`float`
            Area of polygon in steradians.
        """
        if self.gzeroar():
            return 0.0
        warnings.warn(("The ManglePolygon.garea() method is incomplete " +
                       "and is returning a dummy value."),
                      PydlutilsUserWarning)
        return 1.0

    def gzeroar(self):
        """If at least one cap has zero area, then the whole polygon
        has zero area.

        Returns
        -------
        :class:`bool`
            ``True`` if the area is zero.
        """
        if self.ncaps == 0:
            #
            # Catch the case of an "allsky" polygon.
            #
            return False
        return (self.cm == 0.0).any() or (self.cm <= -2.0).any()

    def copy(self):
        """Return an exact copy of the polygon.

        Returns
        -------
        :class:`ManglePolygon`
            A new polygon object.
        """
        return ManglePolygon(self)

    def add_caps(self, x, cm):
        """Create a new polygon that contains additional caps.  The caps
        are appended to the existing polygon's caps.

        Parameters
        ----------
        x : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
            ``X`` values of the new caps.
        cm : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
            ``CM`` values of the new caps.

        Returns
        -------
        :class:`ManglePolygon`
            A new polygon object.
        """
        ncaps = self.ncaps + cm.size
        newdata = dict()
        newdata['weight'] = self.weight
        newdata['pixel'] = self.pixel
        newdata['id'] = self.id
        newdata['use_caps'] = self.use_caps
        newdata['x'] = np.zeros((ncaps, 3), dtype=self.x.dtype)
        newdata['cm'] = np.zeros((ncaps,), dtype=self.cm.dtype)
        newdata['x'][0:self.ncaps, :] = self.x.copy()
        newdata['cm'][0:self.ncaps] = self.cm.copy()
        newdata['x'][self.ncaps:, :] = x.copy()
        newdata['cm'][self.ncaps:] = cm.copy()
        return ManglePolygon(**newdata)

    def polyn(self, other, n, complement=False):
        """Intersection of a polygon with the `n` th cap of another polygon.

        Parameters
        ----------
        other : :class:`~pydl.pydlutils.mangle.ManglePolygon`
            Polygon containing a cap to intersect the first polygon with.
        n : :class:`int`
            Index of the cap in `other`.
        complement : :class:`bool`, optional
            If ``True``, set the sign of the cm value of `other` to be
            the complement of the original value.

        Returns
        -------
        :class:`~pydl.pydlutils.mangle.ManglePolygon`
            A polygon containing the intersected caps.
        """
        x = other.x[n, :]
        sign = 1.0
        if complement:
            sign = -1.0
        cm = sign * other.cm[n]
        return self.add_caps(x, cm)


class FITS_polygon(fits.FITS_rec):
    """Handle polygons read in from a FITS file.

    This class provides aliases for the columns in a typical FITS polygons
    file.
    """
    _pkey = {'X': 'XCAPS', 'x': 'XCAPS',
             'CM': 'CMCAPS', 'cm': 'CMCAPS',
             'ID': 'IFIELD', 'id': 'IFIELD',
             'ncaps': 'NCAPS',
             'weight': 'WEIGHT',
             'pixel': 'PIXEL',
             'str': 'STR',
             'use_caps': 'USE_CAPS'}
    #
    # Right now, this class is only instantiated by calling .view() on
    # a FITS_rec object, so only __array_finalize__ is needed.
    #
    # def __new__(*args):
    #     self = super(FITS_polygon, self).__new__(*args)
    #     self._caps = None
    #     return self

    def __array_finalize__(self, obj):
        super(FITS_polygon, self).__array_finalize__(obj)

    def __getitem__(self, key):
        if isinstance(key, (str,)):
            if key in self._pkey:
                return super(FITS_polygon, self).__getitem__(self._pkey[key])
        return super(FITS_polygon, self).__getitem__(key)

    def __getattr__(self, key):
        if key in self._pkey:
            return super(FITS_polygon, self).__getattribute__(self._pkey[key])
        raise AttributeError("FITS_polygon has no attribute {0}.".format(key))


def angles_to_x(points, latitude=False):
    """Convert spherical angles to unit Cartesian vectors.

    Parameters
    ----------
    points : :class:`~numpy.ndarray`
        A set of angles in the form phi, theta (in degrees).
    latitude : :class:`bool`, optional
        If ``True``, assume that the angles actually represent longitude,
        latitude, or equivalently, RA, Dec.

    Returns
    -------
    :class:`~numpy.ndarray`
        The corresponding Cartesian vectors.
    """
    npoints, ncol = points.shape
    x = np.zeros((npoints, 3), dtype=points.dtype)
    phi = np.radians(points[:, 0])
    if latitude:
        theta = np.radians(90.0 - points[:, 1])
    else:
        theta = np.radians(points[:, 1])
    st = np.sin(theta)
    x[:, 0] = np.cos(phi) * st
    x[:, 1] = np.sin(phi) * st
    x[:, 2] = np.cos(theta)
    return x


def cap_distance(x, cm, points):
    """Compute the distance from a point to a cap, and also determine
    whether the point is inside or outside the cap.

    Parameters
    ----------
    x : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        ``X`` value of the cap (3-vector).
    cm : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        ``CM`` value of the cap.
    points : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        If `points` is a 3-vector, or set of 3-vectors, then assume the point
        is a Cartesian unit vector.  If `point` is a 2-vector or set
        of 2-vectors, assume the point is RA, Dec.

    Returns
    -------
    :class:`~numpy.ndarray`
        The distance(s) to the point(s) in degrees.  If the distance is
        negative, the point is outside the cap.
    """
    npoints, ncol = points.shape
    if ncol == 2:
        xyz = angles_to_x(points, latitude=True)
    elif ncol == 3:
        xyz = points
    else:
        raise ValueError("Inappropriate shape for point!")
    dotprod = np.dot(xyz, x)
    cdist = np.degrees(np.arccos(1.0 - np.abs(cm)) - np.arccos(dotprod))
    if cm < 0:
        cdist *= -1.0
    return cdist


def circle_cap(radius, points):
    """Construct caps based on input points (directions on the unit sphere).

    Parameters
    ----------
    radius : :class:`float` or :class:`~numpy.ndarray`
        Radii of the caps in degrees.  This will become the ``CM`` value.
    points : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        If `points` is a 3-vector, or set of 3-vectors, then assume the point
        is a Cartesian unit vector.  If `point` is a 2-vector or set
        of 2-vectors, assume the point is RA, Dec.

    Returns
    -------
    :class:`tuple`
        A tuple containing ``X`` and ``CM`` values for the cap.
    """
    npoints, ncol = points.shape
    if ncol == 2:
        x = angles_to_x(points, latitude=True)
    elif ncol == 3:
        x = points.copy()
    else:
        raise ValueError("Inappropriate shape for point!")
    if isinstance(radius, float):
        radius = np.zeros((npoints,), dtype=points.dtype) + radius
    elif radius.shape == ():
        radius = np.zeros((npoints,), dtype=points.dtype) + radius
    elif radius.size == npoints:
        pass
    else:
        raise ValueError("radius does not appear to be the correct shape.")
    cm = 1.0 - np.cos(np.radians(radius))
    return (x, cm)


def is_cap_used(use_caps, i):
    """Returns ``True`` if a cap is used.

    Parameters
    ----------
    use_caps : :class:`int`
        Bit mask indicating which cap is used.
    i : :class:`int`
        Number indicating which cap we are interested in.

    Returns
    -------
    :class:`bool`
        ``True`` if a cap is used.
    """
    return (use_caps & 1 << i) != 0


def is_in_cap(x, cm, points):
    """Are the points in a (single) given cap?

    Parameters
    ----------
    x : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        ``X`` value of the cap (3-vector).
    cm : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        ``CM`` value of the cap.
    points : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        If `points` is a 3-vector, or set of 3-vectors, then assume the point
        is a Cartesian unit vector.  If `point` is a 2-vector or set
        of 2-vectors, assume the point is RA, Dec.

    Returns
    -------
    :class:`~numpy.ndarray`
        A boolean vector giving the result for each point.
    """
    return cap_distance(x, cm, points) >= 0.0


def is_in_polygon(polygon, points, ncaps=0):
    """Are the points in a given (single) polygon?

    Parameters
    ----------
    polygon : :class:`~pydl.pydlutils.mangle.ManglePolygon`
        A polygon object.
    points : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        If `points` is a 3-vector, or set of 3-vectors, then assume the point
        is a Cartesian unit vector.  If `point` is a 2-vector or set
        of 2-vectors, assume the point is RA, Dec.
    ncaps : :class:`int`, optional
        If set, use only the first `ncaps` caps in `polygon`.

    Returns
    -------
    :class:`~numpy.ndarray`
        A boolean vector giving the result for each point.
    """
    npoints, ncol = points.shape
    p = dict()
    pmap = {'ncaps': 'NCAPS', 'use_caps': 'USE_CAPS',
            'x': 'XCAPS', 'cm': 'CMCAPS'}
    for key in pmap:
        try:
            p[key] = getattr(polygon, key)
        except AttributeError:
            p[key] = polygon[pmap[key]]
    usencaps = p['ncaps']
    if ncaps > 0:
        usencaps = min(ncaps, p['ncaps'])
    in_polygon = np.ones((npoints,), dtype=bool)
    for icap in range(usencaps):
        if is_cap_used(p['use_caps'], icap):
            in_polygon &= is_in_cap(p['x'][icap, :], p['cm'][icap], points)
    return in_polygon


def is_in_window(polygons, points, ncaps=0):
    """Check to see if `points` lie within a set of `polygons`.

    Parameters
    ----------
    polygons : :class:`PolygonList` or :class:`FITS_polygon`
        A set of polygons.
    points : :class:`~numpy.ndarray` or :class:`~numpy.recarray`
        If `points` is a 3-vector, or set of 3-vectors, then assume the point
        is a Cartesian unit vector.  If `point` is a 2-vector or set
        of 2-vectors, assume the point is RA, Dec.
    ncaps : :class:`int`, optional
        If set, use only the first `ncaps` caps in `polygon`.  This only
        exists to be passed to :func:`is_in_polygon`.

    Returns
    -------
    :class:`tuple`
        A tuple containing two :class:`~numpy.ndarray`.  First, a boolean
        vector giving the result for each point.  Second, an integer vector
        giving the index of the polygon that contains the point.
    """
    npoints, ncol = points.shape
    npoly = len(polygons)
    in_polygon = np.zeros((npoints, ), dtype=np.int32) - 1
    curr_polygon = 0
    while curr_polygon < npoly:
        indx_not_in = (in_polygon == -1).nonzero()[0]
        if len(indx_not_in) > 0:
            indx_in_curr_polygon = is_in_polygon(polygons[curr_polygon],
                                                 points[indx_not_in],
                                                 ncaps=ncaps)
            if indx_in_curr_polygon.any():
                in_polygon[indx_not_in[indx_in_curr_polygon]] = curr_polygon
        curr_polygon += 1
    return (in_polygon >= 0, in_polygon)


def read_fits_polygons(filename, convert=False):
    """Read a "polygon" format FITS file.

    This function returns a subclass of :class:`~astropy.io.fits.FITS_rec`
    that provides some convenient aliases for the columns of a typical
    FITS polygon file (which might be all upper-case).

    Parameters
    ----------
    filename : :class:`str`
        Name of FITS file to read.
    convert : :class:`bool`, optional
        If ``True``, convert the data to a list of :class:`ManglePolygon`
        objects.  *Caution: This could result in some data being discarded!*

    Returns
    -------
    :class:`~pydl.pydlutils.mangle.FITS_polygon` or :class:`list`
        The data contained in HDU 1 of the FITS file.
    """
    with fits.open(filename, uint=True) as hdulist:
        data = hdulist[1].data
    if convert:
        poly = PolygonList()
        for k in range(data.size):
            poly.append(ManglePolygon(data[k]))
    else:
        poly = data.view(FITS_polygon)
    return poly


def read_mangle_polygons(filename):
    """Read a "polygon" format ASCII file in Mangle's own format.  These
    files typically have extension ``.ply`` or ``.pol``.

    Parameters
    ----------
    filename : :class:`str`
        Name of FITS file to read.

    Returns
    -------
    :class:`~pydl.pydlutils.mangle.PolygonList`
        A list-like object containing
        :class:`~pydl.pydlutils.mangle.ManglePolygon` objects and
        any metadata.
    """
    with open(filename, 'r') as ply:
        # lines = ply.read().split(ply.newlines)
        lines = ply.readlines()
    lines = [l.strip() for l in lines]
    try:
        _ = int(lines[0].strip().split()[0])
    except ValueError:
        raise PydlutilsException(("Invalid first line of {0}!  " +
                                  "Are you sure this is a Mangle " +
                                  "polygon file?").format(filename))
    p_lines = [i for i, l in enumerate(lines) if l.startswith('polygon')]
    header = lines[1:p_lines[0]]
    poly = PolygonList(header=header)
    r1 = re.compile(r'polygon\s+(\d+)\s+\(([^)]+)\):')
    mtypes = {'str': float, 'weight': float, 'pixel': int, 'caps': int}
    for p in p_lines:
        m = r1.match(lines[p])
        g = m.groups()
        pid = int(g[0])
        meta = g[1].strip().split(',')
        m1 = [m.strip().split()[1] for m in meta]
        m0 = [mtypes[m1[i]](m.strip().split()[0]) for i, m in enumerate(meta)]
        metad = dict(zip(m1, m0))
        metad['id'] = pid
        metad['x'] = list()
        metad['cm'] = list()
        for cap in lines[p+1:p+1+metad['caps']]:
            data = [float(d) for d in re.split(r'\s+', cap.strip())]
            metad['x'].append(data[0:3])
            metad['cm'].append(data[-1])
        metad['x'] = np.array(metad['x'])
        assert metad['x'].shape == (metad['caps'], 3)
        metad['cm'] = np.array(metad['cm'])
        assert metad['cm'].shape == (metad['caps'],)
        poly.append(ManglePolygon(**metad))
    return poly


def set_use_caps(polygon, index_list, add=False, tol=1.0e-10,
                 allow_doubles=False, allow_neg_doubles=False):
    """Set the bits in use_caps for a polygon.

    Parameters
    ----------
    polygon : :class:`~pydl.pydlutils.mangle.ManglePolygon`
        A polygon object.
    index_list : array-like
        A list of indices of caps to set in the polygon.  Should be no
        longer, nor contain indices greater than the number of caps
        (``polygon.ncaps``).
    add : :class:`bool`, optional
        If ``True``, don't initialize the use_caps value to zero, use the
        existing value associated with `polygon` instead.
    tol : :class:`float`, optional
        Tolerance used to determine whether two caps are identical.
    allow_doubles : :class:`bool`, optional
        Normally, this routine automatically sets use_caps such that no
        two caps with use_caps set are identical.
    allow_neg_doubles : :class:`bool`, optional
        Normally, two caps that are identical except for the sign of `cm`
        would be set unused.  This inhibits that behaviour.

    Returns
    -------
    :class:`int`
        Value of use_caps.
    """
    if not add:
        polygon.use_caps = 0
    t2 = tol**2
    for i in index_list:
        polygon.use_caps |= (1 << index_list[i])
    if not allow_doubles:
        #
        # Check for doubles
        #
        for i in range(polygon.ncaps):
            if is_cap_used(polygon.use_caps, i):
                for j in range(i+1, polygon.ncaps):
                    if is_cap_used(polygon.use_caps, j):
                        if np.sum((polygon.x[i, :] -
                                   polygon.x[j, :])**2) < t2:
                            if ((np.absolute(polygon.cm[i] -
                                             polygon.cm[j]) < tol) or
                                ((polygon.cm[i] +
                                  polygon.cm[j]) < tol and not allow_neg_doubles)):
                                #
                                # Don't use
                                #
                                polygon.use_caps -= 1 << j
    return polygon.use_caps


def x_to_angles(points, latitude=False):
    """Convert unit Cartesian vectors to spherical angles.

    Parameters
    ----------
    points : :class:`~numpy.ndarray`
        A set of x, y, z coordinates.
    latitude : :class:`bool`, optional
        If ``True``, convert the angles to longitude,
        latitude, or equivalently, RA, Dec.

    Returns
    -------
    :class:`~numpy.ndarray`
        The corresponding spherical angles.
    """
    npoints, ncol = points.shape
    phi = np.degrees(np.arctan2(points[:, 1], points[:, 0]))
    r = (points**2).sum(1)
    theta = np.degrees(np.arccos(points[:, 2]/r))
    if latitude:
        theta = 90.0 - theta
    x = np.zeros((npoints, 2), dtype=points.dtype)
    x[:, 0] = phi
    x[:, 1] = theta
    return x


def _single_polygon(poly):
    """Given certain classes of `poly`, return a
    :class:`ManglePolygon` object.

    This function is used to allow various types of input in functions that
    take a *single* polygon as an argument.

    Parameters
    ----------
    poly : object
        An object containing data that could potentially be converted
        into a :class:`ManglePolygon`.

    Returns
    -------
    ManglePolygon
        The resulting *single* polygon object.

    Raises
    ------
    :exc:`ValueError`
        If the conversion is not possible.
    """
    if isinstance(poly, ManglePolygon):
        return poly
    if isinstance(poly, PolygonList) and len(poly) == 1:
        return poly[0]
    if isinstance(poly, FITS_polygon) and len(poly) == 1:
        return ManglePolygon(poly[0])
    if isinstance(poly, fits.fitsrec.FITS_record):
        return ManglePolygon(poly)
    raise ValueError("Can't convert input into a single polygon!")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the image directory in photoop.
"""
import numpy as np


def sdss_psf_recon(psfield, xpos, ypos, normalize=None, trimdim=None):
    """Reconstruct the PSF at position (`xpos`, `ypos`) in an SDSS field.

    These can be read from the `psField files`_.

    .. _`psField files`: https://data.sdss.org/datamodel/files/PHOTO_REDUX/RERUN/RUN/objcs/CAMCOL/psField.html

    Parameters
    ----------
    psfield : :class:`~astropy.io.fits.fitsrec.FITS_rec`
        A PSF KL-decomposition structure read from a psField file.  This is
        from HDU's 1 through 5 in a psField file, corresponding
        to the five filters *u*, *g*, *r*, *i*, *z*.
    xpos : :class:`int`
        Column position (0-indexed, not 0.5-indexed as PHOTO outputs).
    ypos : :class:`int`
        Row position (0-indexed, not 0.5-indexed as PHOTO outputs).
    normalize : :class:`int` or :class:`float`, optional
        If set, normalize the integral of the image to this value.
    trimdim : :class:`tuple`, optional
        Trimmed dimensions; for example, set to (25, 25) to trim
        the output PSF image to those dimensions.  These dimensions
        must be odd-valued.

    Returns
    -------
    :class:`numpy.ndarray`
        The 2D reconstructed PSF image, typically dimensioned (51, 51).
        The center of the PSF is always the central pixel; this function will
        not apply any sub-pixel shifts.

    Notes
    -----
    The SDSS photo PSF is described as a set of eigen-templates, where the
    mix of these eigen-templates is a simple polynomial function with (x,y)
    position on the CCD.  Typically, there are 4 such 51x51 pixel templates.
    The polynomial functions are typically quadratic in each dimension,
    with no cross-terms.

    The formula is the following, where :math:`i` is the index of row
    polynomial order, :math:`j` is the index of column polynomial order,
    and :math:`k` is the template index:

    .. math::

        a_k &= \\sum_i \\sum_j (0.001 \\times \\mathrm{ROWC})^i \\times (0.001 \\times \\mathrm{COLC})^j \\times [C_k]_{ij} \\\\
        \\mathrm{psfimage} &= \\sum_k a_k \\mathrm{RROWS}_k

    The polynomial terms need not be of the same order for each template.

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.io import fits
    >>> from astropy.utils.data import get_readable_fileobj
    >>> from pydl.photoop.image import sdss_psf_recon
    >>> psfile = ('https://data.sdss.org/sas/dr14/eboss/photo/redux/301/' +
    ...           '3366/objcs/3/psField-003366-3-0110.fit')
    >>> with get_readable_fileobj(psfile, encoding='binary') as psField:  # doctest: +REMOTE_DATA
    ...     with fits.open(psField) as hdulist:
    ...         psf = sdss_psf_recon(hdulist[3].data, 500, 500)
    """
    #
    # Hard-wired scale factor for ypos/xpos coefficients.
    #
    rc_scale = 0.001
    #
    # Assume that the dimensions of each eigen-template are the same.
    #
    rnrow = psfield['RNROW'][0]
    rncol = psfield['RNCOL'][0]
    npix = rnrow * rncol
    #
    # These are the polynomial coefficients as a function of x,y.
    # Only compute these coefficients for the maximum polynomial order in use.
    # In general, this order can be different for each eigen-template.
    #
    nr_max = psfield['nrow_b'].max()
    nc_max = psfield['ncol_b'].max()
    # nb = nrow_b * ncol_b
    coeffs = np.outer(((ypos+0.5) * rc_scale)**np.arange(nr_max), ((xpos+0.5) * rc_scale)**np.arange(nc_max))
    #
    # Reconstruct the image by summing each eigen-template.
    #
    neigen = psfield.shape[0]
    psfimage = np.zeros(psfield['RROWS'][0].shape,
                        dtype=psfield['RROWS'][0].dtype)
    for i in range(neigen):
        psfimage += (psfield['RROWS'][i] *
                     (psfield[i]['c'][0:psfield[0]['nrow_b'], 0:psfield[0]['ncol_b']] *
                      coeffs[0:psfield[0]['nrow_b'], 0:psfield[0]['ncol_b']].T).sum())
    #
    # We have reconstructed the PSF as a vector using all the
    # pixels in psfield.RROWS. So, at the end we trim this vector to only
    # those pixels used, and reform it into a 2-dimensional image.
    #
    if normalize is not None:
        psfimage /= psfimage.sum()
        psfimage *= normalize
    psfimage = psfimage[0:npix].reshape(rncol, rnrow)
    if trimdim is not None:
        psfimage = psfimage[(rncol - trimdim[0])//2:(rncol + trimdim[0])//2,
                            (rnrow - trimdim[1])//2:(rnrow + trimdim[1])//2]
    return psfimage



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage implements functions from the idlspec2d package.
"""
#
# Define this early on so that submodules can use it
#
from .. import PydlException
import astropy.utils.exceptions as aue


class Pydlspec2dException(PydlException):
    """Exceptions raised by :mod:`pydl.pydlspec2d` that don't fit into a
    standard exception class like :exc:`ValueError`.
    """
    pass


class Pydlspec2dUserWarning(aue.AstropyUserWarning):
    """Class for warnings issued by :mod:`pydl.pydlspec2d`.
    """
    pass


__all__ = ['Pydlspec2dException', 'Pydlspec2dUserWarning']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.cooling.
"""
import pytest
import numpy as np
from ..cooling import read_ds_cooling


def test_read_ds_cooling():
    with pytest.raises(ValueError):
        logT, loglambda = read_ds_cooling('m-99.cie')
    logT, logL = read_ds_cooling('m-15.cie')
    assert np.allclose(logT[0:5], np.array([4.0, 4.05, 4.1, 4.15, 4.2]))
    assert np.allclose(logL[0:5], np.array([-26.0, -24.66, -23.52,
                                            -22.62, -22.11]))
    logT = np.array([4.025, 4.125, 4.225, 4.325])
    logT2, logL = read_ds_cooling('m-15.cie', logT)
    assert np.allclose(logT2, logT)
    assert np.allclose(logL, np.array([-25.33, -23.07, -22.04, -22.06]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.photoop.image.
"""
import numpy as np
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from ..image import sdss_psf_recon


def test_sdss_psf_recon():
    psField = get_pkg_data_filename('t/psField-003366-3-0110.fit')
    with fits.open(psField) as hdulist:
        psf = sdss_psf_recon(hdulist[3].data, 600, 500)
    assert np.allclose(psf[23:26, 25],
                       np.array([0.02220068, 0.0738798, 0.11940149]))


def test_sdss_psf_norm():
    psField = get_pkg_data_filename('t/psField-003366-3-0110.fit')
    with fits.open(psField) as hdulist:
        psf = sdss_psf_recon(hdulist[3].data, 600, 500, normalize=100.0)
    assert np.allclose(psf.sum(), 100.0)
    assert np.allclose(psf[23:26, 25],
                       np.array([2.194198, 7.301887, 11.801009]))


def test_sdss_psf_recon_trim():
    psField = get_pkg_data_filename('t/psField-003366-3-0110.fit')
    with fits.open(psField) as hdulist:
        psf = sdss_psf_recon(hdulist[3].data, 600, 500, trimdim=(25, 25))
    # print(psf)
    assert np.allclose(psf[10:13, 12],
                       np.array([0.02220068, 0.0738798, 0.11940149]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the yanny directory in idlutils.

This is a Python library for reading & writing yanny files.

:class:`yanny` is an object-oriented interface to SDSS Parameter files
(a.k.a. "FTCL" or "yanny" files) following these specifications_.
Parameter files typically have and can be recognized by the extension
``.par``.  These files may also be recognized if the first line of the
file is::

    #%yanny

This is not part of the standard specification, but it suffices to
identify, *e.g.*, files that have not yet been written to disk, but only
exist as file objects.

Because Parameter files can contain multiple tables, as well as
metadata, there is no simple, one-to-one correspondence between these
files and, say, an astropy :class:`~astropy.table.Table` object.  Thus,
:class:`yanny` objects are implemented as a subclass of
:class:`~collections.OrderedDict` (to remember the order of keyword-value
pairs), and the actual data are values accessed by keyword.  Still, it is
certainly possible to *write* a table-like object to a yanny file.

Given the caveats above, we have introduced *experimental* support for
reading and writing yanny files directly to/from :class:`~astropy.table.Table`
objects.  Because of the experimental nature of this support, it must be
activated "by hand" by including this snippet in your code::

    from astropy.table import Table
    from astropy.io.registry import (register_identifier, register_reader,
                                     register_writer)
    from pydl.pydlutils.yanny import (is_yanny, read_table_yanny,
                                      write_table_yanny)

    register_identifier('yanny', Table, is_yanny)
    register_reader('yanny', Table, read_table_yanny)
    register_writer('yanny', Table, write_table_yanny)

Currently multidimensional arrays are only supported for type ``char``, and a
close reading of the specifications indicates that multidimensional arrays
were only ever intended to be supported for type ``char``.  So no
multidimensional arrays, sorry.

.. _specifications: https://www.sdss4.org/dr16/software/par/
"""
import re
import os
import datetime
import warnings
from collections import OrderedDict
import numpy as np
from astropy.table import Table
# from astropy.io.registry import register_identifier, register_writer
from . import PydlutilsException, PydlutilsUserWarning


class yanny(OrderedDict):
    """An object interface to a yanny file.

    Create a yanny object using a yanny file, `filename`.  If the file exists,
    it is read, & the dict structure of the object will be basically the
    same as that returned by ``read_yanny()`` in the efftickle package.

    If the file does not exist, or if no filename is given, a blank
    structure is returned.  Other methods allow for subsequent writing
    to the file.

    Parameters
    ----------
    filename : :class:`str` or file-like, optional
        The name of a yanny file or a file-like object representing a
        yanny file.
    raw : :class:`bool`, optional
        If ``True``, data in a yanny file will *not* be converted into
        astropy :class:`~astropy.table.Table` objects, but will instead be
        retained as raw Python lists.

    Attributes
    ----------
    raw : :class:`bool`
        If ``True``, data in a yanny file will *not* be converted into
        astropy :class:`~astropy.table.Table` objects, but will instead be
        retained as raw Python lists.
    filename : :class:`str`
        The name of a yanny parameter file.  If a file-like object was used
        to initialize the object, this will have the value 'in_memory.par'.
    _symbols : :class:`dict`
        A dictionary containing the metadata describing the tables.
    _contents : :class:`str`
        The complete contents of a yanny parameter file.
    _struct_type_caches : :class:`dict`
        A dictionary of dictionaries, one dictionary for every structure
        definition in a yanny parameter file.  Contains the types of
        each column
    _struct_isarray_caches : :class:`dict`
        A dictionary of dictionaries, one dictionary for every structure
        definition in a yanny parameter file.  Contains a boolean value
        for every column.
    _enum_cache : :class:`dict`
        Initially ``None``, this attribute is initialized the first time
        the :meth:`isenum` method is called.  The keyword is the name of the
        enum type, the value is a list of the possible values of that type.

    """

    @staticmethod
    def get_token(string):
        """Removes the first 'word' from string.

        If the 'word' is enclosed in double quotes, it returns the
        contents of the double quotes. If the 'word' is enclosed in
        braces, it returns the contents of the braces, but does not
        attempt to split the array.  If the 'word' is the last word of the
        string, remainder is set equal to the empty string.  This is
        basically a wrapper on some convenient regular expressions.

        Parameters
        ----------
        string : :class:`str`
            A string containing words.

        Returns
        -------
        :class:`tuple`
            A tuple containing the first word and the remainder of the string.

        Examples
        --------
        >>> from pydl.pydlutils.yanny import yanny
        >>> yanny.get_token("The quick brown fox")
        ('The', 'quick brown fox')
        """
        if string[0] == '"':
            (word, remainder) = re.search(r'^"([^"]*)"\s*(.*)',
                                          string).groups()
        elif string[0] == '{':
            (word, remainder) = re.search(r'^\{\s*([^}]*)\s*\}\s*(.*)',
                                          string).groups()
        else:
            try:
                (word, remainder) = re.split(r'\s+', string, 1)
            except ValueError:
                (word, remainder) = (string, '')
        # if remainder is None:
        #     remainder = ''
        return (word, remainder)

    @staticmethod
    def protect(x):
        """Used to appropriately quote string that might contain whitespace.

        This method is mostly for internal use by the yanny object.

        Parameters
        ----------
        x : :class:`str`
            The data to protect.

        Returns
        -------
        :class:`str`
            The data with white space protected by quotes.

        Examples
        --------
        >>> from pydl.pydlutils.yanny import yanny
        >>> yanny.protect('This string contains whitespace.')
        '"This string contains whitespace."'
        >>> yanny.protect('This string contains a #hashtag.')
        '"This string contains a #hashtag."'
        """
        if isinstance(x, np.bytes_):
            s = x.decode()
        else:
            s = str(x)
        if len(s) == 0 or s.find('#') >= 0 or re.search(r'\s+', s) is not None:
            return '"' + s + '"'
        else:
            return s

    @staticmethod
    def trailing_comment(line):
        """Identify a trailing comment and strip it.

        This routine works on the theory that a properly quoted comment mark
        will be surrounted by an odd number of double quotes, & we can
        skip to searching for the last one in the line.

        Parameters
        ----------
        line : :class:`str`
            A line from a yanny file potentially containing trailing comments.

        Returns
        -------
        :class:`str`
            The line with any trailing comment and any residual white space
            trimmed off.

        Notes
        -----
        This may fail in certain pathological cases, for example if a
        real trailing comment contains a single double-quote::

            # a 'pathological" trailing comment

        or if someone is over-enthusiastically commenting::

            # # # # # I like # characters.

        Examples
        --------
        >>> from pydl.pydlutils.yanny import yanny
        >>> yanny.trailing_comment('mystruct 1234 "#hashtag" # a comment.')
        'mystruct 1234 "#hashtag"'
        >>> yanny.trailing_comment('mystruct 1234 "#hashtag" # a "comment".')
        'mystruct 1234 "#hashtag"'
        """
        lastmark = line.rfind('#')
        if lastmark >= 0:
            #
            # Count the number of double quotes in the remainder of the line
            #
            if (len([c for c in line[lastmark:] if c == '"']) % 2) == 0:
                #
                # Even number of quotes
                #
                return line[0:lastmark].rstrip()
        return line

    @staticmethod
    def dtype_to_struct(dt, structname='mystruct', enums=None):
        """Convert a NumPy dtype object describing a record array to
        a typedef struct statement.

        The second argument is the name of the structure.
        If any of the columns are enum types, enums must
        be a dictionary with the keys the column names, and the values
        are a tuple containing the name of the enum type as the first item
        and a tuple or list of possible values as the second item.

        Parameters
        ----------
        dt : :class:`numpy.dtype`
            The dtype of a NumPy record array.
        structname : :class:`str`, optional
            The name to give the structure in the yanny file.  Defaults to
            'MYSTRUCT'.
        enums : :class:`dict`, optional
            A dictionary containing enum information.  See details above.

        Returns
        -------
        :class:`dict`
            A dictionary suitable for setting the 'symbols' dictionary of a
            new yanny object.

        Examples
        --------
        """
        dtmap = {'i2': 'short', 'i4': 'int', 'i8': 'long', 'f4': 'float',
                 'f8': 'double'}
        returnenums = list()
        if enums is None:
            enums = dict()
        else:
            for e in enums:
                lines = list()
                lines.append('typedef enum {')
                for n in enums[e][1]:
                    lines.append("    {0},".format(n))
                lines[-1] = lines[-1].strip(',')
                lines.append('}} {0};'.format(enums[e][0].upper()))
                returnenums.append("\n".join(lines))
                # lines.append('')
        lines = list()
        lines.append('typedef struct {')
        for c in dt.names:
            if dt[c].kind == 'V':
                t = dt[c].subdtype[0].str[1:]
                l = dt[c].subdtype[1][0]
                s = dt[c].subdtype[0].itemsize
            else:
                t = dt[c].str[1:]
                l = 0
                s = dt[c].itemsize
            line = '    '
            if t[0] in 'SU':
                if c in enums:
                    line += enums[c][0].upper()
                else:
                    line += 'char'
            else:
                line += dtmap[t]
            line += ' {0}'.format(c)
            if l > 0:
                line += "[{0:d}]".format(l)
            if t[0] in 'SU' and c not in enums:
                line += "[{0:d}]".format(s)
            line += ';'
            lines.append(line)
        lines.append('}} {0};'.format(structname.upper()))
        return {structname.upper(): list(dt.names),
                'enum': returnenums, 'struct': ["\n".join(lines)]}

    def __init__(self, filename=None, raw=False):
        """Create a yanny object using a yanny file.
        """
        super(yanny, self).__init__()
        #
        # The symbol hash is inherited from the old read_yanny
        #
        self._symbols = dict()
        #
        # Create special attributes that contain the internal status of the
        # object. This should prevent overlap with keywords in the data files.
        #
        self.filename = ''
        self._contents = ''
        #
        # Since the re is expensive, cache the structure types keyed by
        # the field. Create a dictionary for each structure found.
        #
        self._struct_type_caches = dict()
        self._struct_isarray_caches = dict()
        self._enum_cache = None
        #
        # Optionally convert numeric data into NumPy arrays
        #
        self.raw = raw
        #
        # If the file exists, read it
        #
        if filename is not None:
            #
            # Handle file-like objects
            #
            if isinstance(filename, (str,)):
                if os.access(filename, os.R_OK):
                    self.filename = filename
                    with open(filename, 'r') as f:
                        self._contents = f.read()
            else:
                #
                # Assume file-like
                #
                self.filename = 'in_memory.par'
                contents = filename.read()
                if 'b' in filename.mode:
                    contents = contents.decode('ascii')
                self._contents = contents
            self._parse()
        return

    def __str__(self):
        """Implement the ``str()`` function for yanny objects.

        Simply prints the current contents of the yanny file.
        """
        return self._contents

    __repr__ = __str__

    def __eq__(self, other):
        """Test two yanny objects for equality.

        Two yanny objects are assumed to be equal if their contents are equal.
        """
        if isinstance(other, yanny):
            return self._contents == other._contents
        return NotImplemented

    def __ne__(self, other):
        """Test two yanny objects for inequality.

        Two yanny objects are assumed to be unequal if their contents are
        unequal.
        """
        if isinstance(other, yanny):
            return self._contents != other._contents
        return NotImplemented

    def __bool__(self):
        """Give a yanny object a definite truth value.

        A yanny object is considered ``True`` if its contents are non-zero.
        """
        return len(self._contents) > 0

    # `__nonzero__` is needed for Python 2.
    # Python 3 uses `__bool__`.
    # http://stackoverflow.com/a/2233850/498873
    __nonzero__ = __bool__

    def type(self, structure, variable):
        """Returns the type of a variable defined in a structure.

        Returns ``None`` if the structure or the variable is undefined.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column whose type you want.

        Returns
        -------
        :class:`str`
            The type of the variable.
        """
        if structure not in self:
            return None
        if variable not in self.columns(structure):
            return None
        #
        # Added code to cache values to speed up parsing large files.
        # 2009.05.11 / Demitri Muna, NYU
        # Find (or create) the cache for this structure.
        #
        try:
            cache = self._struct_type_caches[structure]
        except KeyError:
            self._struct_type_caches[structure] = dict()
            # cache for one struct type
            cache = self._struct_type_caches[structure]
        #
        # Lookup (or create) the value for this variable
        #
        try:
            var_type = cache[variable]
        except KeyError:
            defl = [x for x in self._symbols['struct']
                    if x.find(structure.lower()) > 0]
            defu = [x for x in self._symbols['struct']
                    if x.find(structure.upper()) > 0]
            if len(defl) != 1 and len(defu) != 1:
                return None
            elif len(defl) == 1:
                definition = defl
            else:
                definition = defu
            typere = re.compile(
                r'(\S+)\s+{0}([\[<].*[\]>]|);'.format(variable))
            (typ, array) = typere.search(definition[0]).groups()
            var_type = typ + array.replace('<', '[').replace('>', ']')
            cache[variable] = var_type
        return var_type

    def basetype(self, structure, variable):
        """Returns the bare type of a variable, stripping off any array
        information.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column whose type you want.

        Returns
        -------
        :class:`str`
            The type of the variable, stripped of array information.
        """
        typ = self.type(structure, variable)
        try:
            return typ[0:typ.index('[')]
        except ValueError:
            return typ

    def isarray(self, structure, variable):
        """Returns ``True`` if the variable is an array type.

        For character types, this means a two-dimensional array,
        *e.g.*: ``char[5][20]``.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column to check for array type.

        Returns
        -------
        :class:`bool`
            ``True`` if the variable is an array.
        """
        try:
            cache = self._struct_isarray_caches[structure]
        except KeyError:
            self._struct_isarray_caches[structure] = dict()
            cache = self._struct_isarray_caches[structure]
        try:
            result = cache[variable]
        except KeyError:
            typ = self.type(structure, variable)
            character_array = re.compile(r'char[\[<]\d*[\]>][\[<]\d*[\]>]')
            if ((character_array.search(typ) is not None) or
                    (typ.find('char') < 0 and (typ.find('[') >= 0 or
                                               typ.find('<') >= 0))):
                cache[variable] = True
            else:
                cache[variable] = False
            result = cache[variable]
        return result

    def isenum(self, structure, variable):
        """Returns true if a variable is an enum type.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column to check for enum type.

        Returns
        -------
        :class:`bool`
            ``True`` if the variable is enum type.
        """
        if self._enum_cache is None:
            self._enum_cache = dict()
            if 'enum' in self._symbols:
                for e in self._symbols['enum']:
                    m = re.search(r'typedef\s+enum\s*\{([^}]+)\}\s*(\w+)\s*;',
                                  e).groups()
                    self._enum_cache[m[1]] = re.split(r',\s*', m[0].strip())
            else:
                return False
        return self.basetype(structure, variable) in self._enum_cache

    def array_length(self, structure, variable):
        """Returns the length of an array type or 1 if the variable is not an
        array.

        For character types, this is the length of a two-dimensional
        array, *e.g.*, ``char[5][20]`` has length 5.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column to check for array length.

        Returns
        -------
        :class:`int`
            The length of the array variable
        """
        if self.isarray(structure, variable):
            typ = self.type(structure, variable)
            return int(typ[typ.index('[')+1:typ.index(']')])
        else:
            return 1

    def char_length(self, structure, variable):
        """Returns the length of a character field.

        *e.g.* ``char[5][20]`` is an array of 5 strings of length 20.
        Returns ``None`` if the variable is not a character type. If the
        length is not specified, *i.e.* ``char[]``, it returns the length of
        the largest string.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column to check for char length.

        Returns
        -------
        :class:`int` or None
            The length of the char variable.
        """
        typ = self.type(structure, variable)
        if typ.find('char') < 0:
            return None
        try:
            return int(typ[typ.rfind('[')+1:typ.rfind(']')])
        except ValueError:
            if self.isarray(structure, variable):
                return max([max([len(x) for x in r])
                            for r in self[structure][variable]])
            else:
                return max([len(x) for x in self[structure][variable]])

    def dtype(self, structure):
        """Returns a NumPy dtype object suitable for describing a table as a
        record array.

        Treats enums as string, which is what the IDL reader does.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure.

        Returns
        -------
        :class:`numpy.dtype`
            A dtype object suitable for describing the yanny structure as a
            record array.
        """
        dt = list()
        dtmap = {'short': 'i2', 'int': 'i4', 'long': 'i8', 'float': 'f',
                 'double': 'd'}
        for c in self.columns(structure):
            typ = self.basetype(structure, c)
            if typ == 'char':
                d = "S{0:d}".format(self.char_length(structure, c))
            elif self.isenum(structure, c):
                d = "S{0:d}".format(max([len(x) for x in
                                         self._enum_cache[typ]]))
            else:
                d = dtmap[typ]
            if self.isarray(structure, c):
                dt.append((str(c), str(d), (self.array_length(structure, c),)))
            else:
                dt.append((str(c), str(d)))
        dt = np.dtype(dt)
        return dt

    def convert(self, structure, variable, value):
        """Converts value into the appropriate (Python) type.

        * ``short`` & ``int`` are converted to Python :class:`int`.
        * ``long`` is converted to Python :class:`long`.
        * ``float`` & ``double`` are converted to Python :class:`float`.
        * Other types are not altered.

        There may be further conversions into NumPy types, but this is the
        first stage.

        Parameters
        ----------
        structure : :class:`str`
            The name of the structure that contains `variable`.
        variable : :class:`str`
            The name of the column undergoing conversion.
        value : :class:`str`
            The value contained in a particular row of `variable`.

        Returns
        -------
        :class:`int`, :class:`long`, :class:`float` or :class:`str`
            `value` converted to a Python numerical type.
        """
        intTypes = set(['short', 'int', 'long'])
        floatTypes = set(['float', 'double'])
        typ = self.basetype(structure, variable)
        if typ in intTypes:
            if self.isarray(structure, variable):
                return [int(v) for v in value]
            else:
                return int(value)
        if typ in floatTypes:
            if self.isarray(structure, variable):
                return [float(v) for v in value]
            else:
                return float(value)
        return value

    def tables(self):
        """Returns a list of all the defined structures.

        This is just the list of keys of the object with the 'internal'
        keys removed.
        """
        foo = list()
        for k in self._symbols.keys():
            if k not in ('struct', 'enum'):
                foo.append(k)
        return foo

    def columns(self, table):
        """Returns an ordered list of column names associated with a
        particular table.

        The order is the same order as they are defined in the yanny file.

        Parameters
        ----------
        table : :class:`str`
            The table whose columns are desired.

        Returns
        -------
        :class:`list`
            The list of column names.
        """
        foo = list()
        if table in self._symbols:
            return self._symbols[table]
        return foo

    def size(self, table):
        """Returns the number of rows in a table.

        Parameters
        ----------
        table : :class:`str`
            The table whose size desired.

        Returns
        -------
        :class:`int`
            The number of rows in `table`.
        """
        foo = self.columns(table)
        return len(self[table][foo[0]])

    def pairs(self):
        """Returns a list of keys to keyword/value pairs.

        Equivalent to doing ``self.keys()``, but with all the data tables &
        other control structures stripped out.
        """
        p = list()
        foo = self.tables()
        for k in self.keys():
            if k not in foo:
                p.append(k)
        return p

    def row(self, table, index):
        """Returns a list containing a single row from a specified table in
        column order.

        If index is out of range, it returns an empty list.

        If the yanny object instance is set up for NumPy record arrays, then
        a single row can be obtained with::

            row0 = par['TABLE'][0]

        Parameters
        ----------
        table : :class:`str`
            The table whose row is desired.
        index : :class:`int`
            The number of the row to return.

        Returns
        -------
        :class:`list`
            A row from `table`.
        """
        datarow = list()
        if table in self and index >= 0 and index < self.size(table):
            for c in self.columns(table):
                datarow.append(self[table][c][index])
        return datarow

    def list_of_dicts(self, table):
        """Construct a list of dictionaries.

        Takes a table from the yanny object and constructs a list object
        containing one row per entry. Each item in the list is a dictionary
        keyed by the struct value names.

        If the yanny object instance is set up for NumPy record arrays, then
        the same functionality can be obtained with::

            foo = par['TABLE'][0]['column']

        Parameters
        ----------
        table : :class:`str`
            The table to convert

        Returns
        -------
        :class:`list`
            A list containing the rows of `table` converted to :class:`dict`.
        """
        return_list = list()
        d = dict()
        # I'm assuming these are in order...
        struct_fields = self.columns(table)
        for i in range(self.size(table)):
            one_row = self.row(table, i)  # one row as a list
            j = 0
            for key in struct_fields:
                d[key] = one_row[j]
                j = j + 1
            return_list.append(dict(d))  # append a new dict (copy of d)
        return return_list

    def new_dict_from_pairs(self):
        """Returns a new dictionary of keyword/value pairs.

        The new dictionary (*i.e.*, not a yanny object) contains the keys
        that :meth:`pairs` returns. There are two reasons this is convenient:

        * the key 'symbols' that is part of the yanny object will not be
          present
        * a simple yanny file can be read with no further processing

        Returns
        -------
        :class:`~collections.OrderedDict`
            A dictionary of the keyword-value pairs that remembers the order
            in which they were defined in the file.

        Examples
        --------

        Read a yanny file and return only the pairs::

            >>> from os.path import dirname
            >>> from pydl.pydlutils.yanny import yanny
            >>> y = yanny(dirname(__file__)+'/tests/t/test.par')
            >>> new_dict = y.new_dict_from_pairs()
            >>> new_dict['mjd']
            '54579'
            >>> new_dict['alpha']
            'beta gamma delta'

        added: Demitri Muna, NYU 2009-04-28
        """
        new_dictionary = OrderedDict()
        for key in self.pairs():
            new_dictionary[key] = self[key]
        return new_dictionary

    def write(self, newfile=None, comments=None):
        """Write a yanny object to a file.

        This assumes that the filename used to create the object was not that
        of a pre-existing file.  If a file of the same name is detected,
        this method will *not* attempt to overwrite it, but will print a
        warning. This also assumes that the special 'symbols' key has been
        properly created.  This will not necessarily make the file very
        human-readable, especially if the data lines are long. If the name of a
        new file is given, it will write to the new file (assuming it doesn't
        exist). If the writing is successful, the data in the object will be
        updated.

        Parameters
        ----------
        newfile : :class:`str`, optional
            The name of the file to write.
        comments : :class:`str` or :class:`list` of :class:`str`, optional
            Comments that will be placed at the head of the file.  If a
            single string is passed, it will be written out verbatim, although
            a '#' character will be added if it does not already have one.
            If a list of strings is passed, comment characters will be added
            and the strings will be joined together.
        """
        if newfile is None:
            if len(self.filename) > 0:
                newfile = self.filename
            else:
                raise ValueError("No filename specified!")
        if os.access(newfile, os.F_OK):
            raise PydlutilsException(
                  "{0} exists, aborting write!".format(newfile))
        if comments is None:
            basefile = os.path.basename(newfile)
            timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
            comments = f"""#
# {basefile}
#
# Created by pydl.pydlutils.yanny.yanny
#
# {timestamp}
#
"""
        else:
            if isinstance(comments, (str,)):
                if not comments.startswith('#'):
                    comments = '# ' + comments
                if not comments.endswith('\n'):
                    comments += '\n'
            else:
                comments = ("\n".join(["# {0}".format(c) for c in comments]) +
                            "\n")
        contents = "#%yanny\n" + comments
        #
        # Print any key/value pairs
        #
        for key in self.pairs():
            contents += "{0} {1}\n".format(key, self[key])
        #
        # Print out enum definitions
        #
        if len(self._symbols['enum']) > 0:
            contents += "\n" + "\n\n".join(self._symbols['enum']) + "\n"
        #
        # Print out structure definitions
        #
        if len(self._symbols['struct']) > 0:
            contents += "\n" + "\n\n".join(self._symbols['struct']) + "\n"
        contents += "\n"
        #
        # Print out the data tables
        #
        for sym in self.tables():
            columns = self.columns(sym)
            for k in range(self.size(sym)):
                line = list()
                line.append(sym)
                for col in columns:
                    if self.isarray(sym, col):
                        datum = ('{' + ' '.join([self.protect(x)
                                 for x in self[sym][col][k]]) + '}')
                    else:
                        datum = self.protect(self[sym][col][k])
                    line.append(datum)
                contents += "{0}\n".format(' '.join(line))
        #
        # Actually write the data to file
        #
        with open(newfile, 'w') as f:
            f.write(contents)
        self._contents = contents
        self.filename = newfile
        self._parse()
        return

    def append(self, datatable):
        """Appends data to an existing FTCL/yanny file.

        Tries as much as possible to preserve the ordering & format of the
        original file.  The datatable should adhere to the format of the
        yanny object, but it is not necessary to reproduce the 'symbols'
        dictionary.  It will not try to append data to a file that does not
        exist.  If the append is successful, the data in the object will be
        updated.

        Parameters
        ----------
        datatable : :class:`dict`
            The data to append.
        """
        if len(self.filename) == 0:
            raise ValueError("No filename is set for this object. " +
                             "Use the filename attribute to set the filename!")
        if not isinstance(datatable, dict):
            raise ValueError("Data to append is not of the correct type. " +
                             "Use a dict!")
        timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
        contents = ''
        #
        # Print any key/value pairs
        #
        for key in datatable.keys():
            if key.upper() in self.tables() or key == 'symbols':
                continue
            contents += "{0} {1}\n".format(key, datatable[key])
        #
        # Print out the data tables
        #
        for sym in self.tables():
            if sym.lower() in datatable:
                datasym = sym.lower()
            else:
                datasym = sym
            if datasym in datatable:
                columns = self.columns(sym)
                for k in range(len(datatable[datasym][columns[0]])):
                    line = list()
                    line.append(sym)
                    for col in columns:
                        if self.isarray(sym, col):
                            datum = ('{' + ' '.join([self.protect(x)
                                     for x in datatable[datasym][col][k]]) +
                                     '}')
                        else:
                            datum = self.protect(datatable[datasym][col][k])
                        line.append(datum)
                    contents += "{0}\n".format(' '.join(line))
        #
        # Actually write the data to file
        #
        if len(contents) > 0:
            contents = ("# Appended by yanny.py at {0}.\n".format(timestamp) +
                        contents)
            if os.access(self.filename, os.W_OK):
                with open(self.filename, 'a') as f:
                    f.write(contents)
                self._contents += contents
                self._parse()
            else:
                raise PydlutilsException(self.filename +
                                         " does not exist, aborting append!")
        else:
            warnings.warn("Nothing to be appended!", PydlutilsUserWarning)
        return

    def _parse(self):
        """Converts text into tables that users can use.

        This method is for use internally by the yanny object.  It is not
        meant to be called by users.

        Parsing proceeds in this order:

        #. Lines that end with a backslash character are reattached
           to following lines.
        #. Structure & enum definitions are identified, saved into the
           'symbols' dictionary & stripped from the contents.
        #. Structure definitions are interpreted.
        #. At this point, the remaining lines of the original file can only
           contain these things:

           * 'blank' lines, including lines that only contain comments
           * keyword/value pairs
           * structure rows

        #. The remaining lines are scanned sequentially.

           #. 'Blank' lines are identified & ignored.
           #. Whitespace & comments are stripped from non-blank lines.
           #. Empty double braces ``{{}}`` are converted into empty double
              quotes ``""``.
           #. If the first word on a line matches the name of a structure,
              the line is broken up into tokens & each token or set of tokens
              (for arrays) is converted to the appropriate Python type.
           #. If the first word on a line does not match the name of a
              structure, it must be a keyword, so this line is interpreted
              as a keyword/value pair.  No further processing is done to
              the value.

        #. At the conclusion of parsing, if ``self.raw`` is ``False``, the
           structures are converted into NumPy record arrays.
        """
        #
        # there are five things we might find
        # 1. 'blank' lines including comments
        # 2. keyword/value pairs (which may have trailing comments)
        # 3. enumeration definitions
        # 4. structure definitions
        # 5. data
        #
        lines = self._contents
        #
        # Reattach lines ending with \
        #
        lines = re.sub(r'\\\s*\n', ' ', lines)
        #
        # Find structure & enumeration definitions & strip them out
        #
        self._symbols['struct'] = re.findall(r'typedef\s+struct\s*\{[^}]+\}\s*\w+\s*;',
                                             lines)
        self._symbols['enum'] = re.findall(r'typedef\s+enum\s*\{[^}]+\}\s*\w+\s*;',
                                           lines)
        lines = re.sub(r'typedef\s+struct\s*\{[^}]+\}\s*\w+\s*;', '', lines)
        lines = re.sub(r'typedef\s+enum\s*\{[^}]+\}\s*\w+\s*;', '', lines)
        #
        # Interpret the structure definitions
        #
        typedefre = re.compile(r'typedef\s+struct\s*\{([^}]+)\}\s*(\w*)\s*;')
        for typedef in self._symbols['struct']:
            typedefm = typedefre.search(typedef)
            (definition, name) = typedefm.groups()
            self[name.upper()] = dict()
            self._symbols[name.upper()] = list()
            definitions = re.findall(r'\S+\s+\S+;', definition)
            for d in definitions:
                d = d.replace(';', '')
                (datatype, column) = re.split(r'\s+', d)
                column = re.sub(r'[\[<].*[\]>]$', '', column)
                self._symbols[name.upper()].append(column)
                self[name.upper()][column] = list()
        # Remove lines containing only comments
        comments = re.compile(r'^\s*#')
        # Remove lines containing only whitespace
        blanks = re.compile(r'^\s*$')
        #
        # Remove trailing comments, but not if they are enclosed in quotes.
        #
        # trailing_comments = re.compile(r'\s*\#.*$')
        # trailing_comments = re.compile(r'\s*\#[^"]+$')
        #
        # Double empty braces get replaced with empty quotes
        #
        double_braces = re.compile(r'\{\s*\{\s*\}\s*\}')
        if len(lines) > 0:
            for line in lines.split('\n'):
                if len(line) == 0:
                    continue
                if comments.search(line) is not None:
                    continue
                if blanks.search(line) is not None:
                    continue
                #
                # Remove leading & trailing blanks & comments
                #
                line = line.strip()
                line = self.trailing_comment(line)
                # line = trailing_comments.sub('',line)
                line = double_braces.sub('""', line)
                #
                # Now if the first word on the line does not match a
                # structure definition it is a keyword/value pair
                #
                (key, value) = self.get_token(line)
                uckey = key.upper()
                if uckey in self._symbols:
                    #
                    # Structure data
                    #
                    for column in self._symbols[uckey]:
                        if len(value) > 0 and blanks.search(value) is None:
                            (data, value) = self.get_token(value)
                            if self.isarray(uckey, column):
                                #
                                # An array value
                                # if it's character data, it won't be delimited
                                # by {} unless it is a multidimensional
                                # string array.  It may or may not be delimited
                                # by double quotes
                                #
                                # Note, we're assuming here that the only
                                # multidimensional arrays are string arrays
                                #
                                arraydata = list()
                                while len(data) > 0:
                                    (token, data) = self.get_token(data)
                                    arraydata.append(token)
                                self[uckey][column].append(
                                    self.convert(uckey, column, arraydata))
                            else:
                                #
                                # A single value
                                #
                                self[uckey][column].append(
                                    self.convert(uckey, column, data))
                        else:
                            break
                else:
                    #
                    # Keyword/value pair
                    #
                    self[key] = value
        #
        # If self.raw is False, convert tables into NumPy record arrays
        #
        if not self.raw:
            for t in self.tables():
                record = np.zeros((self.size(t),), dtype=self.dtype(t))
                for c in self.columns(t):
                    record[c] = self[t][c]
                self[t] = record.view(np.recarray)
        return


def write_ndarray_to_yanny(filename, datatables, structnames=None,
                           enums=None, hdr=None, comments=None):
    """Converts a NumPy record array into a new FTCL/yanny file.

    Returns a new yanny object corresponding to the file.

    Parameters
    ----------
    filename : :class:`str`
        The name of a parameter file.
    datatables : :class:`numpy.ndarray`, :class:`numpy.recarray`, :class:`list`
        A NumPy record array containing data that can be copied into a
        `yanny` object. A :class:`list` *containing* arrays can also be passed.
    structnames : :class:`str` or :class:`list` of :class:`str`, optional
        The name(s) to give the structure(s) in the yanny file.  Defaults to
        'MYSTRUCT0'.
    enums : :class:`dict`, optional
        A dictionary containing enum information.  See the documentation for
        the :meth:`~pydl.pydlutils.yanny.yanny.dtype_to_struct` method of the
        yanny object.
    hdr : :class:`dict`, optional
        A dictionary containing keyword/value pairs for the 'header' of the
        yanny file.
    comments : :class:`str` or :class:`list` of :class:`str`, optional
        A string containing comments that will be added to the start of the
        new file.

    Returns
    -------
    :class:`yanny`
        The `yanny` object resulting from writing the file.

    Raises
    ------
    PydlutilsException
        If `filename` already exists, or if the metadata are incorrect.
    """
    par = yanny(filename)
    if par:
        #
        # If the file already exists
        #
        raise PydlutilsException(
              "Apparently {0} already exists.".format(filename))
    if isinstance(datatables, (np.ndarray, np.recarray, Table)):
        datatables = (datatables,)
    if structnames is None:
        structnames = ["MYSTRUCT{0:d}".format(k)
                       for k in range(len(datatables))]
    if isinstance(structnames, (str,)):
        structnames = (structnames,)
    if len(datatables) != len(structnames):
        raise PydlutilsException(
              "The data tables and their names do not match!")
    for k in range(len(datatables)):
        struct = par.dtype_to_struct(datatables[k].dtype,
                                     structname=structnames[k], enums=enums)
        par._symbols['struct'] += struct['struct']
        par._symbols[structnames[k].upper()] = struct[structnames[k].upper()]
        if enums is not None and len(par._symbols['enum']) == 0:
            par._symbols['enum'] = struct['enum']
        par[structnames[k].upper()] = datatables[k]
    if hdr is not None:
        for key in hdr:
            par[key] = hdr[key]
    par.write(filename, comments=comments)
    return par


def is_yanny(origin, path, fileobj, *args, **kwargs):
    """Identifies Yanny files or objects.

    This function is for use with
    :func:`~astropy.io.registry.register_identifier`.

    Parameters
    ----------
    origin : :class:`str`
        'read' or 'write'
    path : :class:`str`
        Path to the file.
    fileobj : file object
        Open file object, if available.

    Returns
    -------
    :class:`bool`
        ``True`` if the file or object is a Yanny file.
    """
    if fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            signature = fileobj.read(7)
        finally:
            fileobj.seek(loc)
        return signature == b'#%yanny'
    elif path is not None:
        return path.endswith('.par')
    return isinstance(args[0], yanny)


def read_table_yanny(filename, tablename=None):
    """Read a yanny file into a :class:`~astropy.table.Table`.

    Because yanny files can contain multiple tables, it is necessary
    to specify the table name to return.  However, all "headers"
    (keyword-value pairs) will be included in the Table metadata.

    This function is for use with
    :func:`~astropy.io.registry.register_reader`.

    Parameters
    ----------
    filename : :class:`str`
        Name of the file to read.
    tablename : :class:`str`
        The name of the table to read from the file.

    Returns
    -------
    :class:`~astropy.table.Table`
        The table read from the file.

    Raises
    ------
    :exc:`PydlutilsException`
        If `tablename` is not set.
    :exc:`KeyError`
        If `tablename` does not exist in the file.
    """
    if tablename is None:
        raise PydlutilsException("The name of the table is required!")
    #
    # When opened by Table.read(), the filename will actually be a file-like
    # object opened in *binary* mode.
    #
    par = yanny(filename)
    try:
        t0 = par[tablename.upper()]
    except KeyError:
        raise KeyError("No table named {0} in {1}!".format(tablename, filename))
    t = Table(t0)
    t.meta = par.new_dict_from_pairs()
    return t


def write_table_yanny(table, filename, tablename=None, overwrite=False):
    """Write a :class:`~astropy.table.Table` to a yanny file.

    If `table` has any metadata, it will be written to the file as well.

    This function is for use with
    :func:`~astropy.io.registry.register_writer`.

    Parameters
    ----------
    table : :class:`astropy.table.Table`
        The object to be written.
    filename : :class:`str`
        Name of the file to write to.
    tablename : :class:`str`, optional
        Name to give `table` within the file.
    overwrite : :class:`bool`, optional
        If ``True``, any existing file will be silently overwritten.
    """
    if overwrite and os.path.exists(filename):
        os.remove(filename)
    if table.meta:
        hdr = table.meta
    else:
        hdr = None
    write_ndarray_to_yanny(filename, table, structnames=tablename,
                           hdr=hdr, comments='Table')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the coord directory in idlutils.
"""
import numpy as np
import astropy.units as u
import astropy.coordinates as ac
from ..goddard.astro import get_juldate


class SDSSMuNu(ac.BaseCoordinateFrame):
    """SDSS Great Circle Coordinates

    Attributes
    ----------
    stripe
        SDSS `Stripe Number`_ .
    node
        Node of the great circle with respect to the celestial equator.
        In SDSS, this is almost always RA = 95.0 degrees.
    incl
        Inclination of the great circle with respect to the celestial
        equator.
    phi
        Counter-clockwise position angle w.r.t. north for an arc
        in the +nu direction.

    Parameters
    ----------
    mu : :class:`~astropy.coordinates.Angle`
        Angle corresponding to longitude measured along a stripe.
    nu : :class:`~astropy.coordinates.Angle`
        Angle corresponding to latitude measured perpendicular to a stripe.

    Notes
    -----
    https://www.sdss4.org/dr16/algorithms/surveycoords/

    .. _`Stripe Number`: https://www.sdss4.org/dr16/help/glossary/#stripe
    """
    default_representation = ac.SphericalRepresentation
    frame_specific_representation_info = {
        'spherical': [ac.RepresentationMapping(reprname='lon', framename='mu',
                                               defaultunit=u.deg),
                      ac.RepresentationMapping(reprname='lat', framename='nu',
                                               defaultunit=u.deg)]}
    frame_specific_representation_info['unitspherical'] = frame_specific_representation_info['spherical']
    stripe = ac.Attribute(default=0)
    node = ac.QuantityAttribute(default=ac.Angle(95.0, unit=u.deg),
                                unit=u.deg)
    # phi = ac.QuantityFrameAttribute(default=None, unit=u.deg)

    @property
    def incl(self):
        return ac.Angle(stripe_to_incl(self.stripe), unit=u.deg)


def current_mjd():
    """Return the current MJD.
    """
    return get_juldate() - 2400000.5


@ac.frame_transform_graph.transform(ac.FunctionTransform, SDSSMuNu, ac.ICRS)
def munu_to_radec(munu, icrs_frame):
    """Convert from SDSS great circle coordinates to equatorial coordinates.

    Parameters
    ----------
    munu : :class:`~pydl.pydlutils.coord.SDSSMuNu`
        SDSS great circle coordinates (mu, nu).

    Returns
    -------
    :class:`~astropy.coordinates.ICRS`
        Equatorial coordinates (RA, Dec).
    """
    # from pydlutils.coord import stripe_to_eta
    # from pydlutils.goddard.misc import cirrange
    # if 'stripe' in kwargs:
    #     node = 95.0
    #     incl = stripe_to_incl(kwargs['stripe'])
    # elif 'node' in kwargs and 'incl' in kwargs:
    #     node = kwargs['node']
    #     incl = kwargs['incl']
    # else:
    #     raise ValueError('Must specify either STRIPE or NODE,INCL!')
    # if mu.size != nu.size:
    #     raise ValueError('Number of elements in MU and NU must agree!')
    sinnu = np.sin(munu.nu.to(u.radian).value)
    cosnu = np.cos(munu.nu.to(u.radian).value)
    sini = np.sin(munu.incl.to(u.radian).value)
    cosi = np.cos(munu.incl.to(u.radian).value)
    sinmu = np.sin((munu.mu - munu.node).to(u.radian).value)
    cosmu = np.cos((munu.mu - munu.node).to(u.radian).value)
    xx = cosmu * cosnu
    yy = sinmu * cosnu * cosi - sinnu * sini
    zz = sinmu * cosnu * sini + sinnu * cosi
    ra = ac.Angle(np.arctan2(yy, xx), unit=u.radian) + munu.node
    dec = ac.Angle(np.arcsin(zz), unit=u.radian)
    # if 'phi' in kwargs:
    #     phi = np.rad2deg(np.arctan2(cosmu * sini,
    #         (-sinmu * sinnu * sini + cosnu * cosi)*cosnu))
    #     return (ra, dec, phi)
    # else:
    #     return (ra, dec)
    return ac.ICRS(ra=ra, dec=dec).transform_to(icrs_frame)


@ac.frame_transform_graph.transform(ac.FunctionTransform, ac.ICRS, SDSSMuNu)
def radec_to_munu(icrs_frame, munu):
    """Convert from equatorial coordinates to SDSS great circle coordinates.

    Parameters
    ----------
    icrs_frame : :class:`~astropy.coordinates.ICRS`
        Equatorial coordinates (RA, Dec).

    Returns
    -------
    :class:`~pydl.pydlutils.coord.SDSSMuNu`
        SDSS great circle coordinates (mu, nu).
    """
    # from pydlutils.coord import stripe_to_eta
    # from pydlutils.goddard.misc import cirrange
    # if 'stripe' in kwargs:
    #     node = 95.0
    #     incl = stripe_to_incl(kwargs['stripe'])
    # elif 'node' in kwargs and 'incl' in kwargs:
    #     node = kwargs['node']
    #     incl = kwargs['incl']
    # else:
    #     raise ValueError('Must specify either STRIPE or NODE,INCL!')
    # if ra.size != dec.size:
    #     raise ValueError('Number of elements in RA and DEC must agree!')
    sinra = np.sin((icrs_frame.ra - munu.node).to(u.radian).value)
    cosra = np.cos((icrs_frame.ra - munu.node).to(u.radian).value)
    sindec = np.sin(icrs_frame.dec.to(u.radian).value)
    cosdec = np.cos(icrs_frame.dec.to(u.radian).value)
    sini = np.sin(munu.incl.to(u.radian).value)
    cosi = np.cos(munu.incl.to(u.radian).value)
    x1 = cosdec * cosra
    y1 = cosdec * sinra
    z1 = sindec
    x2 = x1
    y2 = y1 * cosi + z1 * sini
    z2 = -y1 * sini + z1 * cosi
    mu = ac.Angle(np.arctan2(y2, x2), unit=u.radian) + munu.node
    nu = ac.Angle(np.arcsin(z2), unit=u.radian)
    # if 'phi' in kwargs:
    #     sinnu = np.sin(np.deg2rad(nu))
    #     cosnu = np.cos(np.deg2rad(nu))
    #     sinmu = np.sin(np.deg2rad(mu-node))
    #     cosmu = np.cos(np.deg2rad(mu-node))
    #     phi = np.rad2deg(np.arctan2(cosmu * sini,
    #         (-sinmu * sinnu * sini + cosnu * cosi)*cosnu))
    #     return (ra, dec, phi)
    # else:
    #     return (ra, dec)
    return SDSSMuNu(mu=mu, nu=nu, stripe=munu.stripe)


def stripe_to_eta(stripe):
    """Convert from SDSS great circle coordinates to equatorial coordinates.

    Parameters
    ----------
    stripe : :class:`int` or :class:`numpy.ndarray`
        SDSS Stripe number.

    Returns
    -------
    :class:`float` or :class:`numpy.ndarray`
        The eta value in the SDSS (lambda,eta) coordinate system.
    """
    stripe_sep = 2.5
    eta = stripe * stripe_sep - 57.5
    if stripe > 46:
        eta -= 180.0
    return eta


def stripe_to_incl(stripe):
    """Convert from SDSS stripe number to an inclination relative to the
    equator.

    Parameters
    ----------
    stripe : :class:`int` or :class:`numpy.ndarray`
        SDSS Stripe number.

    Returns
    -------
    :class:`float` or :class:`numpy.ndarray`
        Inclination of the stripe relative to the equator (Dec = 0).
    """
    dec_center = 32.5
    eta_center = stripe_to_eta(stripe)
    incl = eta_center + dec_center
    return incl
